# ESA (C) 2000-2021
# 
# This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
#
#    SAS is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    SAS is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with SAS.  If not, see <http://www.gnu.org/licenses/>.

# pyrgsspecplot.py

from .version import VERSION
from pysas import SAS_RELEASE, SAS_AKA

__version__ = f'pyrgsspecplot (pyrgsspecplot-{VERSION}) [{SAS_RELEASE}-{SAS_AKA}]' 

# pyrgsspecplot


import numpy as np
import pysas.pyutils.pyutils as pyutils
import pysas.pysasplot_utils.pysasplot_utils as sasplt
import sys
import pickle
from astropy.io import fits
from astropy.table import Table
import time
import os
import warnings
#warnings.filterwarnings("error")
import glob
import re
from logger import get_logger
from astropy import units as u
import matplotlib.pyplot as plt
from astropy import constants as const
from astropy.coordinates import Angle


logger = get_logger('pyrgsspecplot')


##################################
######### PLOTTING UTILS #########
##################################

def plot_spectrum(spec_data, axisunits, fits_info, t_labels, output, plot_info, beta_calc_info):
    """
    Plots the given spectrum passed as a table object.

    Args:
        spec_data: the table containing the data to be plotted.
        axisunits: the units to plot for the x-axis.
        fits_info: a list containing information to add to the plot.
        t_labels: the labels from the table, related to spec_data.
        output: the name of the plot file.
        plot_info: list or tuple containing basic information to show in the plots.
        beta_calc_info: information needed for the operations between Channel and wavelength.

    Output:
        1 when finished.
    """
    
    logger.debug('Starting graph processing...')
   
    if dark_mode:
        plt.style.use('dark_background')
    else:
        plt.style.use('default')

    fig, ax1 = plt.subplots()
    t_xlabel, t_ylabel, t_elabel = t_labels

    logger.debug('Getting passed data.')
    
    try:
        spec_data, errors_t = spec_data
        logger.debug('Rebinned - separated tables for error and data.')
    except ValueError:
        logger.debug('Only one table for error and data.')
        errors_t = spec_data

    spec_pix, original_min_channel, original_channel_range, spec_val, spec_delta = beta_calc_info
    spec_data[t_elabel][np.isnan(spec_data[t_elabel])] = 0
    ax1.set_ylabel('COUNTS/CHANNEL')
    ax1.plot(spec_data[t_xlabel], spec_data[t_ylabel], label = 'Counts/channel')

    ax1.plot(errors_t[t_xlabel], errors_t[t_elabel], label = 'Error')
    ax1.legend()

    obs_id, exp_id, telescope, instr, date_obs, date_end, object_fits, ra, dec, exp_idstr, order, exposuret, src_label, spectrumtype, srcid = fits_info

    text = 'XMM. {0}. OBJECT: {1} RA: {2}. DEC: {3}\n'.format(instr, object_fits, ra, dec)
    text = text + 'OBS-ID: {0}. EXP-ID: {1}. Exp. Time: {2}.\nSource label: {3}. Source ID: {4}.'.format(obs_id, exp_id, exposuret, src_label, srcid)
    text = text + '\n\n'
    
    rebin_tag, alpha0, d, offaxis = plot_info

    fig.text(0.5, 0.8, text)
    ax1.set_position([0.1, 0.1, 0.8, 0.6])
    text_l = 'DATE-OBS: {0}. \nDATE-END: {1}.\n{2}.\n\n'.format(date_obs, date_end, rebin_tag)
    fig.text(0.1, 0.8, text_l)
    
    logger.debug('{} {}', np.nanmin(spec_data[t_xlabel].data), np.nanmax(spec_data[t_xlabel].data))
    ax1.set_xlim(np.nanmin(spec_data[t_xlabel].data), np.nanmax(spec_data[t_xlabel].data))

    if axisunits == 'rad':
        ax1.set_xlabel('BETA CHANNEL')
        ax1.set_xlim((0, 3000))
        # draws an horizontal linee at x = 0
        if dark_mode:
            ax1.axhline(y = 0, color = 'white')
        else:
            ax1.axhline(y = 0, color = 'black')

        ax2 = ax1.twiny()
        ax2.set_position([0.1, 0.1, 0.8, 0.6])
        ax2.spines['top'].set_position(('axes', 1))
        ticks = ax1.get_xticks()
        ticks[np.where(ticks == 0)] = 1
        lambda_val = beta2lambda(d, alpha0, ticks, order * (-1), offaxis[0])
        
        print('lambda_val', lambda_val)
        lamb_array = np.arange(39) + 1
        lamb_major = np.array((5, 10, 15, 20, 25, 30, 35, 40))
        lamb_array_beta = lambda2beta(d, alpha0, lamb_array, order * (-1), offaxis[0])
        lamb_major_beta = lambda2beta(d, alpha0, lamb_major, order * (-1), offaxis[0])
        betamin = betaref + 0.5 * betawid + original_min_channel * betawid
        print('lambda_major:', lamb_major_beta)
        betarange = betawid * original_channel_range
        majorpos =(lamb_major_beta - betamin) / betarange
        labelpos = (lamb_array_beta - betamin) / betarange
        print('betamin', betamin, 'betarange', betarange)

        print('amjorpos', majorpos)
        #ax2.set_xticks(majorpos)
        ax2.xaxis.set_ticks(majorpos)
        ax2.xaxis.set_ticklabels(lamb_major)
        ax2.xaxis.set_ticks(labelpos, minor = True)
        #ax2.set_xlim(min(lambda_val), max(lambda_val))
        ax2.set_xlabel(r"$\lambda / \AA$")

        # energy position in the plot
        energy_major = np.array((5, 3.5, 2.5, 2, 1.5, 1, 0.5))
        lambda_energy_major = 12.3985 / energy_major
        energy_values = np.linspace(4.7, 0.4, 100)
        lambda_energy_values = 12.3985 / energy_values
        energy_major_pos = (lambda_energy_major - betamin) / betarange
        energy_values_pos = (lambda_energy_values - betamin) / betarange
        ax3 = ax1.twiny()
        ax3.set_position([0.1, 0.1, 0.8, 0.6])
        print('energy major pos:', energy_major_pos)
        ax3.spines['top'].set_position(('axes', 1.1))
        ax3.xaxis.set_ticks(energy_major_pos)
        ax3.xaxis.set_ticklabels(energy_major)
        ax3.xaxis.set_ticks(energy_values_pos, minor = True)
        ax3.set_xlabel('Energy / KeV')

    elif axisunits == 'angstrom':
        ax1.set_xlabel(r"$\lambda / \AA$")
        if dark_mode:
            ax1.axhline(y = 0, color = 'white')
        else:
            ax1.axhline(y = 0, color = 'black')
        #ax2.set_position([0.1, 0.1, 0.8, 0.6])
        ticks = ax1.get_xticks()
        lamb_marks = (5, 10, 15, 20, 25, 30, 35, 40)
        lamb_values = betaref + ticks * betawid
        lambda_values = np.arange(39) + 1
        lambda_pos = []
        for lamb_tick in lambda_values:
            lambda_pos.append(min(lamb_values, key=lambda x:abs(x - lamb_tick)))
        ax1.xaxis.set_ticklabels(np.around(lamb_values, 2))
        ticks[np.where(ticks == 0)] = 1
        energy_val = 12.3985 / lamb_values
        #ax2.xaxis.set_ticks(ticks)
        #ticks = ax1.get_xticks()
        #ticks[np.where(ticks == 0)] = 1
        #lambda_val = beta2lambda(d, alpha0, ticks, order * (-1), offaxis[0])

        #lamb_array = np.arange(39) + 1
        #lamb_major = np.array((5, 10, 15, 20, 25, 30, 35, 40))
        #lamb_array_beta = lambda2beta(d, alpha0, lamb_array, order * (-1), offaxis[0])
        #lamb_major_beta = lambda2beta(d, alpha0, lamb_major, order * (-1), offaxis[0])
        #lambda_pos = []
        #lambda_pos_minor = []

        #for lamb_tick in lamb_major_beta:
        #    lambda_pos.append(min(spec_data[t_xlabel], key=lambda x:abs(x - lamb_tick)))
        #for lamb_tick in lamb_array_beta:
        #    lambda_pos_minor.append(min(spec_data[t_xlabel], key=lambda x:abs(x - lamb_tick)))

        #lambda_pos = np.array(lambda_pos)
        #lambda_pos_minor = np.array(lambda_pos_minor)
        #betamin = betaref + 0.5 * betawid + original_min_channel * betawid
        #betarange = betawid * original_channel_range
        #majorpos = (lambda_pos - betamin) / betarange
        #minorpos = (lambda_pos_minor - betamin) / betarange
        #ax1.set_xticks(majorpos)
        #ax1.xaxis.set_ticks(majorpos)
        #ax1.xaxis.set_ticklabels(lamb_major)
        #ax1.xaxis.set_ticks(minorpos, minor = True)
        #ax2 = ax1.twiny()

        #ax2.set_xlim(min(lambda_val), max(lambda_val))
        #ax2.set_xlabel('Energy / KeV')

    fig.set_size_inches(12, 9)
    
    fig.suptitle('pyrgsspecplot -- {0} SPECTRUM; ORDER {1}'.format(spectrumtype, order))
    
    if out_format.upper() == 'XS':
        if os.getenv['DISPLAY']:
            plt.show()
        else:
            logger.warning('Display not available.') 
    else:
        try:
            logger.debug('Saving {}.{}.', output, out_format)
            plt.savefig(output + '.' + out_format)
        except FileNotFoundError:
            try:
                dirs, rfile = os.path.split(output + '.' + out_format)
                os.makedirs(dirs)
                plt.savefig(output + '.' + out_format)
            except:
                logger.error('Could not create {}. Could not resolve the given path.', output + '.' + out_format)


##################################
########### MISC UTILS ###########
##################################

def rebin_min_counts(table, min_counts, xcol, ycol, errcol = None):
    """
    Easy tool for 2D rebinning.
    
    Args:
        table: teh table to be rebinned.
        min_counts: the minimum counts per bin.
        xcol: name of the column with the x data.
        ycol: name of the column with the y counts.
        errcol: none by default. The name of the column with the errors.
    
    Output:
        t: the rebinned table.
    """

    logger.debug('Running rebbin_min_counts.')
    t = table
    newcounts = [0] * len(t)
    newchannel = [0] * len(t)
    newyerror = [0] * len(t)
    low_bin = [t[0][xcol]]
    high_bin = []

    howfar = 0 # how much of the table has been processed so far
    k = 0 # positions in the new tabble
    looped = 0 # number of rows a single new bin has needed to reach min_counts

    while howfar < len(t):
        howfar = howfar + looped
        used_bins = 0 # numbins
        runtotal = 0
        looped = 0

        for i in range(howfar, len(t)):
            if runtotal > min_counts:
                low_bin.append(i)
                high_bin.append(i)
                break
            used_bins = used_bins + 1
            if np.isnan(t[i][ycol]):
                t[i][ycol] = 0
                t[i][errcol] = 0
            runtotal = runtotal + t[i][ycol]
            newchannel[k] = newchannel[k] + t[i][xcol]
            newcounts[k] = newcounts[k] + t[i][ycol]
            looped = looped + 1
       
        if k < len(t) - 1:
            try:
                newchannel[k] = newchannel[k] / used_bins
                newcounts[k] = newcounts[k] / used_bins
            except ZeroDivisionError:
                newchannel[k] = newchannel[k] / 1
                newcounts[k] = newcounts[k] / 1
            newyerror[k] = np.sqrt(abs(newcounts[k]))
            k = k + 1

    #if axisunits == 'angstrom':
     #   for j in range(0, len(newchannel)):
      #      logger.debug('angs values: {} {} {}'.format(j, newchannel[j], (betaref + betawid * newchannel[j]) / order))
       #     newchannel[j] = (betaref + betawid * newchannel[j]) / order
        

#    logger.debug('i data: {} {}'.format(len(low_bin), low_bin))
    newchannel = newchannel[0:k-3]
    newcounts = newcounts[0:k-3]
    newyerror = newyerror[0:k-3]
    ts = Table((newchannel, newcounts, newyerror), names = (xcol, ycol, errcol))
    
    #if axisunits == 'rad':
    #    for j in range(0, len(newchannel)):
    #        newchannel[j] = (spec_val - spec_pix * spec_delta - 0.5 * spec_delta) + newchannel[j] * spec_delta
    
#    logger.debug('total K: {}'.format(ts))
    
    
    copy_newe = []
    for i in newyerror:
        copy_newe.extend([i, i])

    low_bin.append(high_bin[-1])

    copy_newy = []
    for i in newcounts:
        copy_newy.extend([i, i])

    #creating a 'binning' grid for the x-axis of the plots
    pos = [item for sublist in zip(low_bin, high_bin) for item in sublist]
    
    #if axisunits == 'angstrom':
     #   for j in range(0, len(pos)):
      #      pos[j] = (betaref + betawid * pos[j]) / order


    new_t = Table([pos[:-2], copy_newy, copy_newe], names = (xcol, ycol, errcol))

    return new_t, ts


def get_more_recent_ccf(abspaths):
    """
    Returns the absolute path of the most recent CCF file.

    Args:
        abspaths: the absolute path of the files to evaluate.

    Output:
        the absolute path of the most recent file.
    """

    maxn = 0
    for f in abspaths:
        f_aux = os.path.basename(f)
        num = int(re.search(r'\d+', f_aux).group(0))
        if num > maxn:
            maxn = num
            maxf = f

    return maxf


def check_xmm_misc(spec_header):
    """
    Checks the XMM miscellanous data CCF file to search for RGS values.

    Args:
        spec_header: the header of the spectrum.
        instr: the name of the instrument.

    Output:
        (d, alpha0): the line separation and the main angle.
    """

    logger.debug('Looking for XMM misc data...')
    
    instr = spec_header['INSTRUME']

    d = 0
    alpha0 = 0

    try:
        alpha0 = spec_header['RGSALPHA']
    except KeyError:
        alpha0 = 0
    try:
        d = spec_header['LINE_SEP']
    except KeyError:
        d = 0

    if d != 0 and alpha0 != 0:
        return d, alpha0

    try:
        ccfpath = os.environ['SAS_CCFPATH']
    except KeyError:
        logger.error('Could not locate variable SAS_CCFPATH. Quitting...')
        sys.exit(0)

    folders = ccfpath.split(':')

    if len(folders) == 1:
        xmm_ccf_file = glob.glob(os.environ['SAS_CCFPATH'] + '/XMM_MISCDATA*')
        xmm_ccf_file = get_more_recent_ccf(xmm_ccf_file)
        logger.debug('CCF file used: {}.', xmm_ccf_file)
        with fits.open(xmm_ccf_file) as xmm:
            try:
                alpha0_data = xmm[1].data[xmm[1].data['PARM_ID'] == 'INCIDENCE_ANGLE'] 
                alpha0 = alpha0_data[alpha0_data['INSTRUMENT_ID'] == instr.upper()]['PARM_VAL'][0]
                alpha0_unit = alpha0_data[alpha0_data['INSTRUMENT_ID'] == instr.upper()]['PARM_UNIT'][0]
                if alpha0_unit.lower() == 'deg':
                    alpha0 = alpha0 * 180 / np.pi
                d_data = xmm[1].data[xmm[1].data['PARM_ID'] == 'GRAT_LINE_DENS']
                d = d_data[d_data['INSTRUMENT_ID'] == instr.upper()]['PARM_VAL'][0]
            except (KeyError, IndexError) as e:
                logger.warning('Could not locate alpha0 and d in the XMM CCF file.')
    else:
        xmm_ccf_list = []
        logger.debug('List of folders found in CCF path: {}', folders)
        for folder in folders:
            try:
                xmm_ccf_file = os.path.abspath(glob.glob(folder + '/XMM_MISCDATA*')[0])
            except IndexError:
                logger.debug('{} does not contain XMM misc. data calibration file.', folder)
                continue
            xmm_ccf_list.append(xmm_ccf_file)
        logger.debug('Found the following XMM misc. data CCF compatible files: {}', xmm_ccf_list)
        latest_ccf = get_more_recent_ccf(xmm_ccf_list)
        logger.debug('CCF file used: {}.', latest_ccf)

        with fits.open(latest_ccf) as xmm:
            try:
                alpha0_data = xmm[1].data[xmm[1].data['PARM_ID'] == 'INCIDENCE_ANGLE']
                alpha0 = alpha0_data[alpha0_data['INSTRUMENT_ID'] == instr.upper()]['PARM_VAL'][0]
                alpha0_unit = alpha0_data[alpha0_data['INSTRUMENT_ID'] == instr.upper()]['PARM_UNIT'][0]
                if alpha0_unit.lower() == 'deg':
                    logger.debug('Original alpha in degrees')
                    alpha0 = alpha0 * np.pi / 180
                d_data = xmm[1].data[xmm[1].data['PARM_ID'] == 'GRAT_LINE_DENS']
                d = d_data[d_data['INSTRUMENT_ID'] == instr.upper()]['PARM_VAL'][0]
            except (KeyError, IndexError) as e:
                logger.warning('Could not locate d and alpha0 in the XMM CCF file.')

    if alpha0 == 0 or d == 0:
        logger.error('Could not locate d or alpha0 neither in the Spectrum file or the XMM CCF file.')
        sys.exit(0)

    logger.info('Line separation and alpha angle retrieved from ccf file. Values: {}, {}.', alpha0, d)
    return(float(d), float(alpha0))
    

def beta2lambda(d, alpha0, beta_values, order, offaxis):
    """
    Return the lambda equivalence to the beta channel array in the input.

    Args:
        d: the line separation of the spectrometer.
        alpha0: the angle of the spectrometer.
        lambda_values: the array containing the channel values.
        order: the order of the spectrum.
        offaxis: the value from the axis.

    Output:
        lamb_values: the lambda equivalence to the beta channel input.
    """

    if order == 0:
        logger.warning('Cannot obtain lambda values with order 0')
        return np.zeros(len(beta_values))

    lamb_values = (np.cos(beta_values) - np.cos(alpha0 + offaxis)) * (d / order)

    return lamb_values


def lambda2beta(d, alpha0, lambda_values, order, offaxis):
    """
    Transforms the beta channel values into wavelength measurements.

    Args:
        d: the line separation of the spectrometer.
        alpha0: the angle of the spectrometer.
        lambda_values: the array containing the channel values.
        order: the order of the spectrum.
        offaxis: the value from the axis.

    Output:
        beta: the beta conversion.
    """
    
    cosbeta = np.cos(alpha0 + offaxis) + order * lambda_values / d
    
    print('    cosbeta = np.cos(alpha0 + offaxis) + order * lambda_values / d', alpha0, offaxis, order, lambda_values, d)
    if sasplt.is_iterable(cosbeta):
        for i in range(0, len(cosbeta)):
            if cosbeta[i] > 1:
                cosbeta[i] = 0
    else:
        if cosbeta > 1:
            cosbeta = 0

    beta = np.arccos(cosbeta)

    return beta


def run(iparsdic):
    print(f'Executing {__file__} {iparsdic}')
    t_start = time.time()

    group = iparsdic['group']
    if 'T' in group.upper() or 'Y' in group.upper():
        group = True
    else:
        group = False

    global dark_mode
    dark_mode = iparsdic['dark_mode']
    if 'Y' in dark_mode.upper() or 'T' in dark_mode.upper():
        dark_mode = True
    else:
        dark_mode = False

    global out_format

    device = iparsdic['device']
    device = device.replace('/', '')

    srclist = iparsdic['sourcelistset']
    sourceid = int(iparsdic['sourceid'])
    
    plotfile = iparsdic['plotfile']
    plotfile = os.path.abspath(plotfile)
    plotfile, out_format = os.path.splitext(plotfile)


    if out_format == '':
        out_format = device
    else:
        if out_format.upper() != device.upper():
            logger.warning('Mismatch between device parameter and the extension given in the plotfile parameter. {} will be used.', device)
        out_format = device

#    if out_format == '':
#        out_format = 'pdf'
#        logger.warning('Could not find format. Using {0}.{1} as default.'.format(plotfile, out_format))
#    else:
#        out_format = out_format.replace('.', '')
#        if not out_format.upper() in ('PDF', 'PNG'):
#            logger.error('Format {} not supported. Please use pdf or png.'.format(out_format))

    if out_format.upper() == 'PS':
        out_format = 'pdf'
        logger.warning('PostScript format devaluated. Using PDF instead.')

    format_flag = sasplt.check_format_compatibility(out_format)
    if format_flag:
        pass
    else:
        logger.warning('Format ({}) not supported. Using PDF...', out_format)
        out_format = 'pdf'

    global rebin
    rebin = iparsdic['rebin']

    spectrumsets = iparsdic['spectrumsets']
    if ',' in spectrumsets:
        spectrumsets = spectrumsets.replace(' ', '')
        spectrumsets = spectrumsets.split(',')
    else:
        spectrumsets = spectrumsets.split(' ')

    nspectrumsets = len(spectrumsets)
    if nspectrumsets == 0:
        logger.error('Empty list of spectra..')
        sys.exit(0)

    if 'T' in rebin.upper() or 'Y' in rebin.upper():
        rebin = True
    else:
        rebin = False

    global mincounts
    mincounts = int(iparsdic['mincounts'])
    if mincounts < 0:
        logger.error('mincounts must be greater than zero.')
        sys.exit(0)

    nspec = 0

    for spectrum in spectrumsets:
        if nspectrumsets == 1:
            pass
        else:
            nspec = nspec + 1
            
        logger.info('Evaluating file {}.', spectrum)
        with fits.open(spectrum) as f:
            spec_data = f[1].data
            logger.debug('Filtering data using pyutils...')
            spec_data = Table(spec_data)
            # commented out in order to use nan in the binning
            #spec_data = pyutils.filter_data(spec_data) 
            spec_header = f[1].header

        try:
            hduclas2 = spec_header['HDUCLAS2']
        except KeyError:
            logger.warning('Keyword HDUCLAS2 missing in the spectrum file {}. Selecting BACKGROUND as default.', spectrum)
            spectrumtype = 'BACKGROUND'
        if hduclas2 == 'TOTAL':
            spectrumtype = 'SRC+BKG'
        elif hduclas2 == 'NET':
            spectrumtype = 'NET'
        else:
            spectrumtype = 'BACKGROUND'

        global axisunits
        try:
            tcuni = spec_header['TCUNI']
        except KeyError:
            try:
                tcuni = spec_header['TCUNI1']
            except KeyError:
                logger.warning('Keyword TCUNI missing in the spectrum file {}. Selecting rad as default.', spectrum)
                tcuni = 'rad'
                axisunits = 'BACKGROUND'
        
        if tcuni == 'Angstrom':
            axisunits = 'angstrom'
        elif tcuni == 'rad':
            axisunits = 'rad'
        else:
            axisunits = 'rad'

        src_label, offaxis = get_src_label_and_offaxes(srclist, sourceid)
        exposuret = float(pyutils.get_key_word(spectrum, 'Exposure'))
        global order
        order = spec_header['RFLORDER']

        global spec_delta
        try:
            spec_delta = spec_header['SPECDELT']
        except KeyError:
            logger.warning('SPECDELT not found')
            spec_delta = 1
        global spec_pix
        try:
            spec_pix = spec_header['SPECPIX']
        except KeyError:
            logger.warning('SPECPIX not found')
            spec_pix = 0
        global spec_val
        try:
            spec_val = spec_header['SPECVAL']
        except KeyError:
            logger.warning('SPECVAL not found')
            spec_val = 0

        detchans = spec_header['DETCHANS']
        nbins = len(spec_data)

        global betaref
        try:
            betaref = spec_header['TCRVL']
        except KeyError:
            try:
                betaref = spec_header['TCRVL1']
            except KeyError:
                logger.error('Could not locate reference in the input file.')
                sys.exit(0)
        global betawid
        try:
            betawid = spec_header['TCDLT']
        except KeyError:
            try:
                betawid = spec_header['TCDLT1']
            except KeyError:
                logger.error('Could not locate width reference in the input file.')
                sys.exit(0)

        channel_p = spec_data['CHANNEL']
        xmin = sorted(channel_p)[0] - (sorted(channel_p)[2] - sorted(channel_p)[1]) / 2.0
        xmax = sorted(channel_p)[-1] + (sorted(channel_p)[-1] - sorted(channel_p)[-2]) / 2.0

        try:
            counts_error_p = spec_data['STAT_ERR']
        except KeyError:
            pass

        t_elabel = 'STAT_ERR'
        t_xlabel = 'CHANNEL'

        try:
            counts_p = spec_data['COUNTS']
            t_ylabel = 'COUNTS'
            try:
                yerror = spec_data['STAT_ERR']
            except KeyError:
                yerror = np.sqrt(abs(counts_p))
        except KeyError:
            logger.debug('RATE-like spectrum...')
            try:
                rate_p = spec_data['RATE']
                try:
                    counts_p = exposuret * rate_p
                except RuntimeError:
                    logger.error('{} {}', exposuret, rate_p)
                    raise RuntimeError
                try:
                    yerror = np.sqrt(abs(counts_p))
                except KeyError:
                    yerror = np.sqrt(abs(spec_data['RATE'] * exposuret))
                t_ylabel = 'RATE'
            except KeyError:
                logger.error('Could not detect either COUNTS or RATE columns in spectrum data.')
                sys.exit(0)
        
        yerror = np.sqrt(abs(counts_p))

        data_table = Table((channel_p, counts_p, yerror), names = (t_xlabel, t_ylabel, t_elabel))
        
        # Grouping analysis...
        if group:
            try:
                grouping = spec_data['GROUPING']
                grouping_flag = True
            except KeyError:
                logger.warning('Could not locate the GROUPING column in file {}.', spectrum)
                grouping_flag = False

            if grouping_flag:
                logger.info('Working on grouped spectrum.')
                total_new_bins = 0
                bin_c = 0
                
                for i in range(0, nbins):
                    if grouping[i] == 1:
                        total_new_bins = total_new_bins + 1

                summed_counts = [0] * total_new_bins
                n_channels = [0] * total_new_bins
                summed_channels = [0] * total_new_bins
                
                for i in range(1, nbins):
                    if t_ylabel == 'COUNTS':
                        summed_counts[bin_c] = summed_counts[bin_c] + counts_p[i - 1]
                    elif t_ylabel == 'RATE':
                        summed_counts[bin_c] = summed_counts[bin_c] + counts_p[i - 1]
                    try:
                        summed_channels[bin_c] = summed_channels[bin_c] + channel_p[i - 1]
                        n_channels[bin_c] = n_channels[bin_c] + 1
                    except IndexError:
                        logger.error('An error has ocurred while trying to form the grouped channels.')
                        sys.exit(0)

                    if grouping[i] == 1:
                        bin_c = bin_c + 1
                
                channel = []
                counts = []
               
                channel = np.divide(summed_channels, n_channels)
                if np.isnan(channel[-1]):
                    channel[-1] = nbins
                counts = summed_counts
                y_error = np.sqrt(counts)
                data_table = Table((channel, counts, y_error), names = (t_xlabel, t_ylabel, t_elabel))

        # Rebinning...
        if rebin:
            logger.info('Rebinning for file {}.', spectrum)
            original_min_channel = xmin
            original_channel_range = xmax
            #DTin, DTout = sasplt.get_time_deltas(spectrum, 1, mincounts, 'CHANNEL')
            rebin_tag = 'Rebin: Minimum counts per bin: {0}'.format(mincounts)
            tstable = rebin_min_counts(data_table, mincounts, t_xlabel, t_ylabel, t_elabel)
        else:
            original_channel_range = spec_delta * detchans
            original_min_channel = spec_val - spec_pix * spec_delta - 0.5 * spec_delta
            DTin, DTout = 1, 1
            rebin_tag = 'No rebinning'
            tstable = sasplt.TSrebin(data_table, DTin, DTout, t_xlabel, t_ylabel, t_elabel)

        fits_info = collect_fits_info(spectrum)

        d, alpha0 = check_xmm_misc(spec_header)
        plot_info = (rebin_tag, alpha0, d, offaxis)
        fits_info.append(order)
        fits_info.append(exposuret)
        fits_info.append(src_label)
        fits_info.append(spectrumtype)
        fits_info.append(sourceid)
        t_labels = (t_xlabel, t_ylabel, t_elabel)

        beta_calc_info = spec_pix, original_min_channel, original_channel_range, spec_val, spec_delta
        
        if nspec != 0:
            if out_format.upper() == 'PDF':
                save_as = plotfile + '_{0}_TEMP_SPECPLOT'.format(nspec)
            else:
                save_as = plotfile + '_{0}'.format(nspec)
        else:
            save_as = plotfile
        
        plot_spectrum(tstable, axisunits, fits_info, t_labels, save_as, plot_info, beta_calc_info)


    if nspec > 1 and out_format.upper() == 'PDF':
        logger.info('Merging existing PDFs...')
        pdf_files = glob.glob(plotfile + '*_TEMP_SPECPLOT.{}'.format(out_format))
        status = sasplt.merge_pdf(pdf_files, plotfile + '.{}'.format(out_format))

        if status:
            for pdf_f in pdf_files:
                os.remove(pdf_f)
        else:
            logger.warning('Failed to merge the PDFs.')

    t_stop = time.time()

    logger.info('Finished running pyrgsspecplot in {} seconds.', round(t_stop - t_start, 2))


def get_src_label_and_offaxes(srclist, sourceid):
    """
    Returns the label and the offaxis values matching the source id from the source list.

    Args:
        srclist: the FITS file containing the source list info.
        sourceid: the source id that is being evaluated.

    Output:
        src_label: the label taken from the source file.
        offaxis: tuple with the DELTA_DISP and the DELTA_XDSP values.
    """
    
    logger.debug('Getting source label and offaxis values from {} with id {}.', srclist, sourceid)

    with fits.open(srclist) as f:
        try:
            src_label = f['SRCLIST'].data['LABEL'][sourceid - 1]
        except (KeyError, IndexError) as e:
            logger.warning('Could not locate label with the current source id ({}) and source list ({}). Switching to sourceid = 1.', sourceid, srclist)
            src_label = 'Indef'
            sourceid = 1
        try:
            src_delta_disp = f['SRCLIST'].data['DELTA_DISP'][sourceid - 1]
            src_delta_xdsp = f['SRCLIST'].data['DELTA_XDSP'][sourceid - 1]
            src_delta_disp = Angle(src_delta_disp, u.arcminute)
            src_delta_xdsp = Angle(src_delta_xdsp, u.arcminute)

            offaxis = (src_delta_disp.radian, src_delta_xdsp.radian)
            offaxis = (src_delta_disp.value * np.pi / 180 / 60, src_delta_xdsp.value * np.pi / 180 / 60)
        except (KeyError, IndexError) as e:
            logger.warning('Could not locate the offaxis values within the current source list and source id.')
            offaxis = (0,0)

    return src_label, offaxis


def collect_fits_info(specfits):
    """
    Returns the needed information from a spectrum fits file.

    Args:
        specfits: the spectrum fits file, either the path or an already opened
    astropy object.

    Output:
        obs_id, exp_id, telescope, instr, date_obs, date_end, object_fits, ra, dec,
    exp_idstr
    """

    obs_id = pyutils.get_key_word(specfits, 'OBS_ID')
    exp_id = pyutils.get_key_word(specfits, 'EXP_ID')
    telescope = pyutils.get_key_word(specfits, 'TELESCOP')
    instr = pyutils.get_key_word(specfits, 'INSTRUME')
    date_obs = pyutils.get_key_word(specfits, 'DATE-OBS')
    date_end = pyutils.get_key_word(specfits, 'DATE-END')
    object_fits = pyutils.get_key_word(specfits, 'OBJECT')

    if object_fits == 'unknown':
        object_fits = 'Indef'

    ra = pyutils.get_key_word(specfits, 'RA_OBJ')
    if ra == 'unknown':
        ra = pyutils.get_key_word(specfits, 'RA')
        if ra == 'unknown':
            ra = 'Indef'
    try:
        ra = round(ra, 3)
    except TypeError:
        pass

    dec = pyutils.get_key_word(specfits, 'DEC_OBJ')
    
    if dec == 'unknown':
        dec = pyutils.get_key_word(specfits, 'DEC')
        if dec == 'unknown':
            dec = 'Indef'
    
    try:
        dec = round(dec, 3)
    except TypeError:
        pass

    exp_idstr = pyutils.get_key_word(specfits, 'EXPIDSTR')

    return[obs_id, exp_id, telescope, instr, date_obs, date_end, object_fits, ra, dec, exp_idstr]




#!/usr/bin/env python

#	ESA (C) 2000-2018 
#	This file is part of ESA's XMM-Newton Scientific Analysis System
#	(SAS).
#
#	SAS is free software: you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation, either version 3 of the License, or
#	(at your option) any later version.
#
#	SAS is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#	   
#	You should have received a copy of the GNU General Public License
#	along with SAS.  If not, see <http://www.gnu.org/licenses/>.


import sys
import os
import subprocess
import traceback
import threading
import time
import logging

class doIt:
    def __init__(self,fileName):

        
        self.logger = logging.getLogger('xmmextractorGUI')
        self.logger.info("Creating Canvas")
        if len(fileName) == 0:
            self.fileName = ''
        else:
            self.fileName = fileName                
        
        error = ''
        status = None        
        timeout = 100
        thread_count = 2
        self.threads = []
        self.threadName = 'xmmextractor'
        self.job = None
        self.pid = None
        self.jobPIDDict = dict()
        self.stdoutFile=open("xmmextractor.log",'w')
     
    def runSASCommand(self, SASTask):
        #child_env = os.environ.copy()  
        os.remove("xmmextractor.info") if os.path.exists("xmmextractor.info") else None
        
        #stderrFile=open("xmmextractor.error",w)
        try:
            self.logger.info("running command.... "+ SASTask)
            if SASTask == "xmmextractor":
                self.job = subprocess.Popen("xmmextractor paramfile=%s " % self.fileName, \
                                        universal_newlines = True,\
                                        bufsize = -1, \
                                        stdout = self.stdoutFile, \
                                        stderr = subprocess.STDOUT, shell=True)
##                                        env=child_env,\
              
                self.logger.info("NEW PID " + self.job.pid)
                
            else:
                self.job = subprocess.Popen(SASTask, \
                            universal_newlines = True,\
                            bufsize = 64,\
                            stdout = self.stdoutFile, \
                            stderr = subprocess.STDOUT, shell=True)
    
            
            self.job._internal_poll(_deadstate='dead')
            self.pid = self.job.pid
            #print ("NEW PID " + self.pid)
            self.jobPIDDict[ self.pid] = self.job.poll()
            #self.logger.info("done ",self.job.poll())
        except:
            self.error = traceback.format_exc()
            self.status = -1
                    

            
    def checkJobStatus(self, val):
        
        self.logger.debug ("Checking JobStatus.......... "+ str(self.job.poll()) + "  PID " + str(val) + " getPID " + str(self.getPID()) )
      
        if  self.job.poll() == None:
            return True

        return False
      
       
    def deleteJob(self):
        self.logger.info("DELETING JOB...." )
        val = self.job.pid
        self.pid = None
        self.logger.info("JobStatus after KILL.......... "+ str(self.job.poll()) + "  PID " + str(val))

    def getPID(self):
        return self.job.pid

    def closeLogFile(self):
        self.logger.info("Closing log file." )
        self.stdoutFile.close()
        
    def setPID(self,currentPID):
        self.pid = currentPID

    def updatePID(self,pid):
         self.logger.info ("Updating pid "+str(self.job.pid))
         self.jobPIDDict[self.job.pid] =  -1

    def checkJob(self):
        t = threading.Thread(target=checkJobStatus(),name=self.threadName)
        t.setDaemon(True)

        t.start()
        self.logger.info("Thread launched ")
             



# ESA (C) 2000-2021
# 
# This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
#
#    SAS is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    SAS is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with SAS.  If not, see <http://www.gnu.org/licenses/>.

"""wrapper.py

This file is only here for legacy reasons. Its only purpose
is to point to the MyTask class in sastask.py.
"""

# Standard library imports
from warnings import warn

# Third party imports

# Local application imports
from pysas.sastask import MyTask

class Wrapper:
    """Class Wrapper

    This class is only here for legacy reasons. Its only purpose
    is to point to the MyTask class in sastask.py.

    Method run instantiates MyTask and calls 'run' from MyTask.
    """
    def __init__(self, taskname, inargs, 
                 logfilename = None, 
                 tasklogdir  = None,
                 output_to_terminal = True, 
                 output_to_file     = False):
        self.taskname    = taskname
        self.inargs      = inargs
        self.logfilename = logfilename
        self.tasklogdir  = tasklogdir
        self.output_to_terminal = output_to_terminal
        self.output_to_file     = output_to_file
        warn(
             """
             The wrapper class has been depricated. Use MyTask instead.
             ex: 
                 from pysas.sastask import MyTask
             instead of
                 from pysas.wrapper import Wrapper
             """)

    def run(self):
        MyTask(self.taskname, self.inargs, 
               logfilename = self.logfilename, 
               tasklogdir  = self.tasklogdir,
               output_to_terminal = self.output_to_terminal, 
               output_to_file     = self.output_to_file
               ).run()



# ESA (C) 2000-2020
# 
# This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
#
#    SAS is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    SAS is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
#
# pkgmaker.py
#
# (C) 2020 ESA
#
# Script to create the basic structure of any SAS task regardless of 
# being a task based on C++, Fortran, Perl or Python.
#
# The first version of this script was coded in Perl and created mainly
# the directory structure for a new task.
#
# This one, entirely rewriten in Python, does the same but, for Python, 
# it creates also all the software files required to build a real 
# Python package according to the Python Packaging standard 
# (see the Python Packaging User Guide).
#
# This script is aimed to create new SAS packages which include one or
# more tasks. 

# Standard library imports
import sys, os, shutil
from datetime import date

# Third Party imports


# Local app imports
from .version import VERSION
from pysas import SAS_RELEASE, SAS_AKA


__version__ = f'pkgmaker (pkgmaker-{VERSION}) [{SAS_RELEASE}-{SAS_AKA}]'


def run(iparsdic):

    # current date and year in string format
    today = date.today()
    creationyear = today.year
    creationdate = today.strftime('%d-%b-%Y')
    creationdate_chlog = today.strftime('%Y-%m-%d')


    # SAS_DIR and SAS_PATH must be defined

    try:
        os.environ['SAS_DIR']
    except KeyError:
        print ("Your environment does not define SAS_DIR\n")
        sys.exit(1)

    sasdir = os.environ['SAS_DIR']

    try:
        os.environ['SAS_PATH']
    except KeyError:
        print ("Your environment does not define SAS_PATH\n")
        sys.exit(1)

    saspath = os.environ['SAS_PATH'].split(':')

    # Additional check: SAS_DIR must be equal to saspath[1] 

    if ( saspath[1] != sasdir ):
        print(f'Error: pkgmaker: Mismatching paths\n{saspath[1]}\n{sasdir}')
        sys.exit(1)

    # The base of the user´s development tree is saspath[0]
    # Let us call it sasdevbase

    sasdevbase = saspath[0]


    # Directories
    
    # Until now, basic code skeletons for tasks written in C++, Fortran90,
    # and Perl were put in the src directory as module.cc, module.f90 and
    # module.pl. A first attempt to create a basic skeleton for a Python task
    # named module.py, was also placed there. In all cases, these files
    # were simple hints on how to create the basic structure for the main
    # program of the task in the specific language.
    #
    # With Python we are changing that philosophy to provide both, the
    # directory structure and the files required to build and install
    # a real Python package.
    #
    # With pkgmaker we can create 4 types of SAS packages: C++ (cpp),
    # Fortran90 (fortran), Perl (perl) and Python (python).
    # Those packages may include either one or several tasks.
    #
    # Package templates will be all below pkgtemplate
    #
    #    pkgtemplate = os.path.join(sasdir, 'packages/pkgmaker/template')
    #
    # Each type of package, has its own templatedir
    #
    #     templatedir = os.path.join(pkgtemplate, 'python')
    #                   os.path.join(pkgtemplate, 'cpp')
    #                   os.path.join(pkgtemplate, 'fortran')
    #                   os.path.join(pkgtemplate, 'perl')


    # home is the user´s home directory

    home = os.environ['HOME']

    # Below sasdevbase, developers use to have the SAS git clone,
    # typically named 'sas', which, together with sasdevbase forms
    # sasdev. 

    sasdev = os.path.join(home, sasdevbase + '/sas')

    if not os.path.exists(sasdev):
        print(f"""

        Directory {sasdev} does not exist. 

        Possibly your SAS git clone is not named sas. 

        Please create soft link and rerun pkgmaker.
        
        """)
        sys.exit(1)


    # The following chain of directories is based on sasdir
    #
    # pkgmakerdir is the directory of package pkgmaker
    # pkgtemplate is the directory with the pkgmaker templates

    pkgmakerdir = os.path.join(sasdir, 'packages/pkgmaker')
    pkgtemplate = os.path.join(sasdir, 'packages/pkgmaker/template')

    # packagesdir is where new packages will be written

    packagesdir = os.path.join(sasdev, 'packages')


    # Initial values for taskStatus and taskVersion

    pkgstatus = 'Draft'  # Draft/Submitted/Approved
    pkgversion = '0.1'

    # Begin asking information about the new package

    print(f"""
    
    SAS package maker 
    
    Version: {__version__}

    *

    New SAS packages will be placed in: {packagesdir}
    
    """)

    pkgname = input("\n\nEnter the name of the new package: ")

    # Change dir to packagesdir to check whether pkgname exists or not

    os.chdir(packagesdir)

    if not(os.path.exists(pkgname)):
        os.mkdir(pkgname)
        print(f"\nCreated directory for package {pkgname}")
    else:
        print(f"Error: Directory {pkgname} already exists!")
        sys.exit(1)

    # Request type of package

    pkgtype = {
            '1': 'cpp',
            '2': 'fortran',
            '3': 'perl',
            '4': 'python'
            }

    print("""
    
    Select package programming language:
            
            1 - c++       (cpp)
            2 - fortran   (fortran)
            3 - perl      (perl)
            4 - python    (python)
    
    Default 4 - python
    
    """)

    sel = input("Select 1, 2, 3, 4 : ")

    if sel == "":
        sel = '4'

    if sel in '1234':
        templatedir = os.path.join(pkgtemplate, pkgtype[sel])
        print(f"\nPackage {pkgname} is of type {pkgtype[sel]}")
    else:
        print(f"\nError: Unknown option {sel} ")
        sys.exit(1)

    print(f"""
    
    Package {pkgname} can have one or more tasks.
    
    Will give name to one!
    
    """)

    taskname = input(f'\n\n\nEnter a task name (Default: {pkgname} - <Hit ENTER>): ')
    if not taskname:
        taskname = pkgname

    # Extension of the main program
    taskext = {
        '1': '.cc',
            '2': '.f90',
            '3': '.pl',
            '4': '.py'
            }
    # Sign for comments depends on language
    commentsign = {
            '1': '//',
            '2': '!',
            '3': '#',
            '4': '#'
            }


    # Define home for pkgname

    pkgnamedir = os.path.join(packagesdir, pkgname)

    # Change dir to pkgname

    os.chdir(pkgname)


    # Create the package directory structure similar for all packages

    for dirname in ('src', 'config', 'doc', 'test'):
        os.mkdir(dirname)

    # Fills in the VERSION file with the initial version of the package
    with open('VERSION', 'w') as outf:
        string = pkgversion + '\n' # It requires to append an \n
        outf.write(string)

    # Copy DEPEND and DISTRIBUTION from the tasktemplate directory
    # These are common to all packages
    # Perhaps in a future version of this script, we can request to enter
    # the value of DISTRIBUTION
    shutil.copy(os.path.join(pkgtemplate, 'DEPEND'), '.')
    shutil.copy(os.path.join(pkgtemplate, 'DISTRIBUTION'), '.')

    # Package description

    pkgdescription = input("\nEnter a one-line description of the package: ")

    print(f"""
    
    Package {pkgname} type {pkgtype[sel]} will receive version {pkgversion} and status {pkgstatus}. 
    
    You may want to change these in the LaTeX description document 
    
                 doc/{pkgname}_description.tex 
    
    or in the main program source code
    
                 src/{pkgname}/{pkgname}{taskext[sel]}
    
    
    Now, please enter the names of author(s) of the package, their affiliation(s)
    and email address(es).
    
    """
    )

    authors = input(f"\n\nEnter author(s) for {pkgname} separated by commas: ")
    pkgauthors = authors.split(',')
    for k in range(len(pkgauthors)):
        pkgauthors[k] = pkgauthors[k].strip()

    affiliations = ['a']*len(pkgauthors)
    for k in range(len(pkgauthors)):
        affiliations[k] = input(f"\n\nAffiliation for {pkgauthors[k]}: ")

    emails = ['a']*len(pkgauthors)
    for k in range(len(pkgauthors)):
        emails[k] = input(f"\n\ne-mails for {pkgauthors[k]}: ")

    class Authorscard:
        def __init__(self, author, affiliation, email):
            self.author = author
            self.affiliation = affiliation
            self.email = email

        def __str__(self):
            return(f"{self.author} - {self.affiliation} - {self.email}")

    print("\nPackage authors card:\n")

    authorsvcard = '\n' + commentsign[sel] + '\t\t'
    authorsvcard_tex = '\n\t\t'
    vcard = ['a']*len(pkgauthors)
    for k in range(len(pkgauthors)):
        vcard[k] = Authorscard(pkgauthors[k], affiliations[k], emails[k])
        authorsvcard += str(vcard[k]) + '\n' + commentsign[sel] + '\t\t'
        authorsvcard_tex += str(vcard[k]) + '\n\t\t'
    print(authorsvcard)
    print(authorsvcard_tex)

    pkgauthors_str = ''
    emails_str = ''
    for k in range(len(pkgauthors)):
        if (k != (len(pkgauthors) - 1) ):
            pkgauthors_str += pkgauthors[k] + ' - '
            emails_str += emails[k] + ' - '
            continue
        pkgauthors_str += pkgauthors[k]
        emails_str += emails[k]



    # Write the files which are required by all type of packages
    #
    # pkgtemplate/Changelog.in
    #                                  => pkgnamedir/ChangeLog
    # pkgtemplate/Makefile.in
    #                                  => pkgnamedir/Makefile
    #
    # pkgtemplate/doc/package_description.tex.in
    #                                  => pkgnamedir/doc/pkgname_description.tex
    # pkgtemplate/doc/Makefile.in
    #                                  => pkgnamedir/doc/Makefile
    #
    # pkgtemplate/config/Makefile.in
    #                                  => pkgnamedir/config/Makefile
    # pkgtemplate/config/task.par.in
    #                                  => pkgnamedir/config/pkgname.par
    # pkgtemplate/config/pkg.info.in
    #                                  => pkgnamedir/config/pkgname.info
    #
    # pkgtemplate/test/Makefile.in
    #                                  => pkgnamedir/test/Makefile
    # pkgtemplate/test/notest.in
    #                                  => pkgnamedir/test/notest
    #
    #

    # In C/C++ and Perl programs and in TeX/LaTeX documents the symbols
    # { and } are used to delimit main programs, subroutines and typesetting
    # commands, therefore the "in" files can not have them as such because
    # they will conflict with the delimiters of the variable names {variable}.
    # To solve this problem, we define cb and ce as strings containing only
    # those symbols so we can write them to the output file.
    cb = '{'
    ce = '}'


    with open(os.path.join(pkgtemplate, 'ChangeLog.in')) as inf, \
            open((os.path.join(pkgnamedir, 'ChangeLog')), 'w') as outf:
         template = inf.read()
         outf.write(template.format(pkgname=pkgname,
                                    pkgversion=pkgversion,
                                    creationdate_chlog=creationdate_chlog))

    # For Python based packages, the input file is Makefile.python.in
    if sel in ['1', '2', '3']:
        with open(os.path.join(pkgtemplate, 'Makefile.in')) as inf, \
                open((os.path.join(pkgnamedir, 'Makefile')), 'w') as outf:
         template = inf.read()
         outf.write(template.format(creationyear=creationyear))

    fileout = 'doc/' + pkgname + '_description.tex'
    cb = '{'
    ce = '}'
    with open(os.path.join(pkgtemplate, 'doc/package_description.tex.in')) as inf, \
            open((os.path.join(pkgnamedir, fileout)), 'w') as outf:
         template = inf.read()
         outf.write(template.format(cb=cb,
                                    ce=ce,
                                    creationyear=creationyear,
                                    pkgname=pkgname,
                                    creationdate=creationdate,
                                    pkgdescription=pkgdescription,
                                    authorsvcard=authorsvcard_tex,
                                    pkgversion=pkgversion,
                                    pkgstatus=pkgstatus))

    fileout = 'doc/Makefile'
    with open(os.path.join(pkgtemplate, 'doc/Makefile.in')) as inf, \
            open((os.path.join(pkgnamedir, fileout)), 'w') as outf:
         template = inf.read()
         outf.write(template.format(creationyear=creationyear,
                                    pkgname=pkgname))

    fileout = 'config/Makefile'
    with open(os.path.join(pkgtemplate, 'config/Makefile.in')) as inf, \
            open((os.path.join(pkgnamedir, fileout)), 'w') as outf:
         template = inf.read()
         outf.write(template.format(creationyear=creationyear,
                                    taskname=taskname,
                                    pkgname=pkgname))

    fileout = 'config/' + taskname + '.par'
    with open(os.path.join(pkgtemplate, 'config/task.par.in')) as inf, \
            open((os.path.join(pkgnamedir, fileout)), 'w') as outf:
         template = inf.read()
         outf.write(template.format(creationyear=creationyear,
                                    taskname=taskname))

    fileout = 'config/' + taskname + '.info'
    with open(os.path.join(pkgtemplate, 'config/task.info.in')) as inf, \
            open((os.path.join(pkgnamedir, fileout)), 'w') as outf:
         template = inf.read()
         outf.write(template.format(creationyear=creationyear,
                                    taskname=taskname,
                                    pkgdescription=pkgdescription))

    fileout = 'test/Makefile'
    with open(os.path.join(pkgtemplate, 'test/Makefile.in')) as inf, \
            open((os.path.join(pkgnamedir, fileout)), 'w') as outf:
         template = inf.read()
         outf.write(template.format(creationyear=creationyear))

    fileout = 'test/notest'
    with open(os.path.join(pkgtemplate, 'test/notest.in')) as inf, \
            open((os.path.join(pkgnamedir, fileout)), 'w') as outf:
         template = inf.read()
         outf.write(template.format(creationyear=creationyear,
                                    taskname=taskname))
         # Make notest mode 755
         os.chmod(os.path.join(pkgnamedir, fileout),0o755)

    # Write the source code templatefor a task for each type of package

    # 1 - Package type c++

    if (sel == '1'):
        # module.cc.in => module.cc
        taskmodulein = 'module' + taskext[sel] + '.in'
        taskmodule   = taskname + taskext[sel]
        #print(f"                                     \n \
        #        pkgnamedir       = {pkgnamedir}      \n \
        #        taskmodulein     = {taskmodulein}    \n \
        #        taskmodule       = {taskmodule}      \n \
        #        templatedir      = {templatedir}     \n \
        #        creationyear     = {creationyear}    \n \
        #        creationdate     = {creationdate}    \n \
        #        pkgdescription   = {pkgdescription}  \n \
        #        authorsvcard     = {authorsvcard}")
        #

        with open(os.path.join(templatedir, taskmodulein)) as inf, \
                open((os.path.join('src', taskmodule)), 'w') as outf:
            template = inf.read()
            outf.write(template.format(creationyear=creationyear,
                                       taskname=taskname,
                                       creationdate=creationdate,
                                       pkgdescription=pkgdescription,
                                       authorsvcard=authorsvcard,
                                       cb=cb,
                                       ce=ce))

        with open(os.path.join(templatedir, 'Makefile.in')) as inf, \
                open((os.path.join('src', 'Makefile')), 'w') as outf:
            template = inf.read()
            outf.write(template.format(creationyear=creationyear,
                                       taskname=taskname))

    # 2 - Package type fortran
    elif(sel == '2'):
        # module.f90.in => pkgname_mod.f90
        taskname_mod = taskname + '_mod'
        taskmodulein = 'module' + taskext[sel] + '.in'
        taskmodule   = taskname_mod + taskext[sel]
        #print(f"                                     \n \
        #        pkgnamedir       = {pkgnamedir}      \n \
        #        taskmodulein     = {taskmodulein}    \n \
        #        taskmodule       = {taskmodule}      \n \
        #        templatedir      = {templatedir}     \n \
        #        creationyear     = {creationyear}    \n \
        #        creationdate     = {creationdate}    \n \
        #        pkgdescription   = {pkgdescription}  \n \
        #        authorsvcard     = {authorsvcard}")
        #
        with open(os.path.join(templatedir, taskmodulein)) as inf, \
                open((os.path.join('src', taskmodule)), 'w') as outf:
            template = inf.read()
            outf.write(template.format(creationyear=creationyear,
                                       taskname=taskname,
                                       creationdate=creationdate,
                                       pkgdescription=pkgdescription,
                                       authorsvcard=authorsvcard,
                                       taskname_mod=taskname_mod))

        with open(os.path.join(templatedir, 'Makefile.in')) as inf, \
                open((os.path.join('src', 'Makefile')), 'w') as outf:
            template = inf.read()
            outf.write(template.format(creationyear=creationyear,
                                       taskname=taskname,
                                       taskname_mod=taskname_mod))


    # 3 - Package type perl
    elif(sel == '3'):
        # module.pl.in => pkgname.pl
        taskmodulein = 'module' + taskext[sel] + '.in'
        taskmodule   = taskname + taskext[sel]
        #print(f"                                     \n \
        #        pkgnamedir       = {pkgnamedir}      \n \
        #        taskmodulein     = {taskmodulein}    \n \
        #        taskmodule       = {taskmodule}      \n \
        #        templatedir      = {templatedir}     \n \
        #        creationyear     = {creationyear}    \n \
        #        creationdate     = {creationdate}    \n \
        #        pkgdescription   = {pkgdescription}  \n \
        #        authorsvcard     = {authorsvcard}")
        #
        with open(os.path.join(templatedir, taskmodulein)) as inf, \
                open((os.path.join('src', taskmodule)), 'w') as outf:
            template = inf.read()
            outf.write(template.format(creationyear=creationyear,
                                       taskname=taskname,
                                       creationdate=creationdate,
                                       pkgdescription=pkgdescription,
                                       authorsvcard=authorsvcard,
                                       cb=cb,
                                       ce=ce))

        with open(os.path.join(templatedir, 'Makefile.in')) as inf, \
                open((os.path.join('src', 'Makefile')), 'w') as outf:
            template = inf.read()
            outf.write(template.format(creationyear=creationyear,
                                       taskname=taskname))

    # 4 - Package type python
    elif (sel == '4'):

        classname = 'MyTask'

        os.mkdir(os.path.join('src', pkgname))
        os.mkdir(os.path.join('src', pkgname, 'bin'))

        # Replace Makefile in root dir by Makefile.python.in
        with open(os.path.join(pkgtemplate, 'Makefile.python.in')) as inf, \
                open((os.path.join(pkgnamedir, 'Makefile')), 'w') as outf:
            template = inf.read()
            outf.write(template.format(creationyear=creationyear,
                                       pkgname=pkgname))

        # Create README.md from README.md.in
        with open(os.path.join(templatedir, 'README.md.in')) as inf, \
                open((os.path.join(pkgnamedir, 'README.md')), 'w') as outf:
            template = inf.read()
            outf.write(template.format(pkgname=pkgname))

        # Create Makefile from Makefile.taskname.in
        with open(os.path.join(templatedir, 'Makefile.taskname.in')) as inf, \
                open((os.path.join('src', pkgname, 'Makefile')), 'w') as outf:
            template = inf.read()
            outf.write(template.format(creationyear=creationyear))

        # Create Makefile from Makefile.taskname.bin.in
        with open(os.path.join(templatedir, 'Makefile.taskname.bin.in')) as inf, \
                open((os.path.join('src', pkgname, 'bin', 'Makefile')), 'w') as outf:
            template = inf.read()
            outf.write(template.format(creationyear=creationyear,
                                       taskname=taskname))


        # Create setup.py from setup.py.in
        with open(os.path.join(templatedir, 'setup.py.in')) as inf, \
                open('setup.py', 'w') as outf:
                template = inf.read()
                outf.write(template.format(author=pkgauthors_str,
                                           author_email=emails_str,
                                           pkgname=pkgname,
                                           description=pkgdescription,
                                           taskversion=pkgversion,
                                           taskname=taskname))

        # Create src/pkgname/__init__.py
        with open(os.path.join(templatedir, '__init__.py.in')) as inf, \
                open(os.path.join('src', pkgname, '__init__.py'), 'w') as outf:
                template = inf.read()
                outf.write(template.format(taskname=taskname))

        # Create src/pkgname/taskname.py
        with open(os.path.join(templatedir, 'taskname.py.in')) as inf, \
                open(os.path.join('src', pkgname, f'{taskname}.py'), 'w') as outf:
                template = inf.read()
                outf.write(template.format(creationyear=creationyear,
                                           taskname=taskname,
                                           cb=cb,
                                           ce=ce))

        # Create src/pkgname/bin/taskname
        with open(os.path.join(templatedir, 'taskname.in')) as inf, \
                open(os.path.join('src', pkgname, 'bin', taskname), 'w') as outf:
                template = inf.read()
                outf.write(template.format(creationyear=creationyear,
                                           taskname=taskname,
                                           classname=classname))
                os.chmod(os.path.join('src', pkgname, 'bin', taskname), 0o755)

        # Copy template/python/version.py.in and template/python/write_version.py to 
        # src/pkgname
        shutil.copy(os.path.join(templatedir, 'version.py.in'), \
                os.path.join('src', pkgname))
        shutil.copy(os.path.join(templatedir, 'write_version.py'), \
                os.path.join('src', pkgname))
        # Copy specific DEPEND files for Python packages
        shutil.copy(os.path.join(templatedir, 'DEPEND'), '.')
    else:
        sys.exit(1)



# ESA (C) 2000-2021
# 
# This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
#
#    SAS is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    SAS is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
#
"""sastask.py

All stand alone SAS tasks can be called by using MyTask.

As any other SAS task, Python based tasks are used to
perform specific data processing jobs.

The task invocation command may include several parameters
and generic options.

Parameters allow to set or to choose among specific task processing
options. SAS parameter syntax has the form param=value.
Mandatory parameters must be present in the command line.
Optional parameters might have a default value.
All task parameters with their properties
must be defined in a parameter file, which is unique for each task.
Such file provides for each paramater, its name, type (int, real,
boolean, string, etc), options (one or several choices), whether it is
mandatory or not, its default value if none is given, etc.
The parameter file is written in XML format.

Generic options, common to all SAS tasks, allow to handle
several features of the execution environment.
They can be expressed either with a single or with a double "-",
e.g. -v or --version (to show the task version and exit), -V or
--verbosity (to set the level of output verbosity between 0 and 10), etc.
Some of these options modify shell environment variables, e.g.
SAS_VERBOSITY is modified by the value given to --verbosity, etc.

Let be mytask a new SAS task. It will be an instance of MyTask
 
mytask = MyTask('mytask', args)
  
where args is a list that includes all arguments we pass to mytask
in the command line.

The specific Python code for mytask, and possibly other auxiliary Python
code which might be required by mytask to operate, will be placed all
within the directory structure of mytask.

A typical invocation of mytask could be of the form

MyTask('mytask', ['param0=value0','param1=value1', ... [Options]).run()

where param0, param1, etc, are any task parameters defined in
a file named mytask.par, and [Options] are any of the generic options
common to all SAS task. If a parameter, e.g. param0, is defined in
mytask.par as 'mandatory', it must be present.

Of course, the presence of some of the Options, will trigger specific
immediate actions: e.g. the option -v or --version will immediately
show the version of mytask and exit.
 
Arguments of the form param=value can alternate with Options. 
The only requirement is that Options with a value, e.g. -V 4, must be
adjacent.

Methods defined by MyTask (identified so far):

1. readparfile: Read and loads the task parameter file.

2. processargs: Parses arguments and acts accordingly. 

3. run: Executes the task with the proper arguments.

4. printHelp: Prints information about the task, equivalent to
              MyTask('mytask', ['-h']).run()

The readparfile method used class paramXmlInfoReader from
param.py module.

The processargs methods uses the class ParseArgs based on argparse.
"""

# Standard library imports
import os, numbers, subprocess
from importlib import import_module
import importlib.resources
from warnings import warn

# Third party imports

# Local application imports
from pysas.param import paramXmlInfoReader, SASParams
from pysas.parser import ParseArgs
from pysas.logger import get_logger

# Class MyTask
class MyTask:
    """
    The SASTask class replaces the Wrapper class from
    wrapper.py.

    Inputs:
    (required)
        - taskname   : Name of the task to be run.
        - inargs     : Input arguments.
    (optional)
        - logfilename: Designated log file name. Will be used instead
                        of "{taskname}.log". Useful for putting all 
                        output from multiple tasks into the same file.
        - tasklogdir : (default: cwd) Directory where to write the log 
                        file.
                        Priority of defaults for task_logdir
                        1. tasklogdir (passed in to function)
                        2. SAS_TASKLOGDIR (envirnment variable)
                        3. cwd (final default)

        - output_to_terminal : (default: True) Output will be written to 
                                the terminal.
        - output_to_file     : (default: False) Output will be written to 
                                a log file.


    For pySAS v2.0 the inargs has switched to fundamentally being a dictionary.
    If a list is passed in it will be converted into a dictionary.
                                
    In the class initialization, the task name and
    the input args to run it are processed. 
    Task parameters as identified by the '=' sign are 
    separated of task options, to reorder them to 
    avoid conflicts at the time of parsing them.

    The instance method 'readparfile' gets a full
    picture of the task parameter file, receiving
    a bunch of information abouth subparameters and
    their relationship with its predecessors.

    The instance method 'processargs' performs the
    processing of any immediate options and filter out
    the legitimate and mandatory parameters so as
    they can be used in the 'run' instance method.
    """

    def __init__(self, taskname, 
                 inargs = {}, 
                 logfilename = None, 
                 tasklogdir  = None,
                 output_to_terminal = True, 
                 output_to_file     = False,
                 logger = None):
        self.taskname    = taskname
        self.inargs      = inargs
        self.logfilename = logfilename
        self.tasklogdir  = tasklogdir
        self.output_to_terminal = output_to_terminal
        self.output_to_file     = output_to_file
        
        # The logger object assigned here will only be used for log information
        # generated by pySAS. Log information generated by SAS tasks will not 
        # be controled by this logger object.
        # If a logger object was not passed in, then generate one
        if logger is None:
            if logfilename is None:
                # If no logfilename is given, then only write out to terminal
                self.logger = get_logger('MyTask',
                                         toterminal  = self.output_to_terminal)
            else:
                # If a logfilename was given then use that
                self.logger = get_logger('MyTask',
                                         toterminal  = self.output_to_terminal, 
                                         tofile      = self.output_to_file, 
                                         logfilename = self.logfilename,
                                         tasklogdir  = self.tasklogdir)
        else:
            # If a logger object was passed in, then use that.
            self.logger = logger

        # Check if inargs is a 'dict'.
        if isinstance(self.inargs, dict):
            # Insert empty string for options if not present
            if not 'options' in self.inargs.keys():
                self.inargs['options'] = ''
            for key in self.inargs.keys():
                # Safety check to see if a number was passed in.
                if isinstance(self.inargs[key], numbers.Number):
                        self.inargs[key] = str(self.inargs[key])
                # Convert booleans to 'yes' and 'no'
                if self.inargs[key] is bool:
                    if self.inargs[key]:
                        self.inargs[key] = 'yes'
                    else:
                        self.inargs[key] = 'no'

        # Check if inargs is a 'SASParams' dict.
        if isinstance(self.inargs, SASParams):
            argsdic = {}
            # Pull off the parameters that have been modified
            for k, v in self.inargs.inparams.items():
                if v[1]:
                    argsdic[k] = v[0]
            self.inargs = dict(argsdic)
            if not 'options' in self.inargs.keys():
                self.inargs['options'] = ''

        # Check if inargs is a single string. If it is, then split
        # it at all spaces. This is not fool proof!
        # ***STRONGLY*** NOT RECOMMENDED
        # (unless the input is very simple, such as '-h')
        if isinstance(self.inargs, str):
            self.inargs = self.inargs.split()

        # If inargs is a list, pick off all the options, then put
        # all other arguments into key-value pairs in the dictionary.
        if isinstance(self.inargs, list):
            argsdic = {}
            options = []
            for a in self.inargs:
                if '=' in a:
                    k, v = a.split('=', 1)
                    argsdic[k] = v
                else:
                    options.append(a)
            argsdic['options'] = " ".join(options)
            self.inargs = argsdic

    def __repr__(self):
        return f'{self.__class__.__name__}({self.taskname}, {self.inargs})'

    def readparfile(self):
        """
        Reads the parameter file (taskname.par) for the task. Returns 
        various lists and dictionaries.
            - allparams : Dictionary { parameter: {attributes} }
            - mandpar   : List of all mandatory parameters/subparameters
            - mainparams: List of main paramaters
            - parmap    : Dictionary { pname: [ subparameters ...], ...}
            - mandpar_dict: Dictionary, key = mandatory subparamater, 
                            value=its parent parameter
            - rev_mandpar_dict: Dictionary, keys = parent paramater, 
                                value = list of all mandatory subparameters
            - rev_mandpar_string_dict: Dictionary, keys = parent parameter type 'string'
                                       value = list of alternatives
        """
        t = paramXmlInfoReader(self.taskname, logger = self.logger)
        t.xmlParser()
        self.allparams = t.allparams
        self.mandparams = t.mandpar
        self.mainparams = t.mainparams
        self.parmap = t.parmap
        self.mandpar_dict = t.mandpar_dict
        self.rev_mandpar_dict = t.rev_mandpar_dict
        self.rev_mandpar_string_dict = t.rev_mandpar_string_dict

    def processargs(self):
        # Only if options are passed in
        if 'option' in self.inargs.keys():
            self.logger.warning("Input arguments include the key 'option'. Use 'options' (with an 's') instead.")
                    
        self.Exit = False
        if len(self.inargs['options']) > 0:
            pararg = ParseArgs(self.taskname, self.inargs, logger = self.logger)
            pararg.optparser()

            # Execute options which require immediate action
            self.Exit = pararg.exe_options()
            if self.Exit:
                return self.Exit
            
            # Deal with options that set environment variables
            # For non-Python based tasks these will be passed in
            # as part of the command for the subprocess call.
            # For Python based tasks, they will have to be handled
            # in the .run() function.
            env_opts = pararg.env_options()

        # Remove the options from argsdic
        self.argsdic = dict(self.inargs)
        self.argsdic.pop('options',None)

        # 1st check: Whether or not parameters in inargs are defined in the parameter file
        for p in self.argsdic.keys():
            if p.strip() not in self.allparams.keys():
                raise Exception(f'Parameter {p} is not defined in the parameter file')

        # 2nd check: Whether we have mandatory parameters and subparameters
        for k, v in self.rev_mandpar_dict.items():
            if k.strip() in v and k.strip() not in self.argsdic.keys():
                raise Exception(f'Missing, at least, mandatory parameter {k}')
            if k not in v:
                for p in self.argsdic.keys():
                    if p in v:
                        implicit = 'yes'
                        for c in v:
                            if c not in self.argsdic.keys():
                                implicit = 'no'
                                raise Exception(f'Besides {p}, mandatory subparameter {c} must also be present')
                        if implicit == 'yes':
                            defval = self.allparams[k]['default']
                            if self.allparams[k]['type'] == 'string':
                                self.rev_mandpar_string_dict[k].remove(defval)
                                alt = self.rev_mandpar_string_dict[k].pop()
                            elif self.allparams[k]['type'] == 'bool' and defval == 'no':
                                alt = 'yes'
                            elif self.allparams[k]['type'] == 'bool' and defval == 'yes':
                                alt = 'no'
                            newinarg = k + '=' + alt
                            self.argsdic[k] = alt
                            if k in self.argsdic.keys():
                                self.logger.warning(f'No need to include {k}. Assumed {newinarg}')
                            break
                    elif p in v and k not in self.argsdic.keys():
                        raise Exception(f'Mandatory sub-parameter {p} requires {k} be present')

        # If any subparameters of a parent parameter are set in the command
	    # line or the arguments list, the parent can not keep its default
	    # value. This is typical for parent parameters of type 'boolean' whose
	    # default value is 'no' (or 'yes'). In such a case the parent value must 
        # be changed to 'yes' (or 'no'). This is known as 'implicit' behaviour. 
	    # The previous code takes this behaviour in consideration but
	    # only for mandatory parameters. Now it will be applied to all
	    # parameters having subparameters. 
	    # The final effect of this will be visible below when merging 
	    # the parsdic (the dictionary with the default values for 
	    # all the parameters) and the inargs to pass all task  
	    # parameters for execution.

	    # Now compute a dictionary for all implicit parameters. 
	    # keys are the parent parameters having any subparameters and
	    # values are lists with these subparameters as elements.
	    # Neither the keys nor the elements of the lists can be mandatory.

        implicitparams = {}
        for d in self.parmap:
            for k, v in d.items():
                if v != [] and k not in self.mandparams:
                    for sp in v:
                        if sp not in self.mandparams:
                            implicitparams[k] = v

        # At this point in inargs there should be only optional modifiers
        # and true task parameters

        # Given that we know all parameters for the task (self.allparams) with
        # their default values, and the parameters and their values entered 
        # in the command line, let us produce now a single object with all 
        # parameters which we can pass to the module to run.

        # Load defaults with all parameters default values.
        # Use dictionary method 'setdefault' to set the value for a given key;
        # here the key is 'default'.
        # If the value for that key is not defined, fill it with ''.

        defaults = {}
        for a in self.allparams.values():
            defaults[a['id']] = a.setdefault('default', '')

        # 3rd Check: Whether any subparameter is set in self.argsdic or not.
        # We assume the parent is boolean ('yes'/'no').

        for a in self.argsdic.keys():
            for k, v in implicitparams.items():
                if a in v:
                    #print(f'parent = {k}')
                    if self.allparams[k]['default'] == 'no':
                        defaults[k] = 'yes'
                    elif self.allparams[k]['default']  == 'yes':
                        defaults[k] = 'no'
                    #print(k, a,  defaults[k])
                    break

        # Return options that modify the environment to argdic
        self.env_options = {}
        if len(self.inargs['options']) > 0:
            for k, v in env_opts.items():
                if k == 'env_options':
                    self.argsdic['options'] = env_opts['env_options']
                else:
                    self.env_options[k] = v

        # Merge self.argsdic onto parsdic. Those values set in command line, 
        # from self.argsdic, will overwrite the defaults obtained from the par file, 
        # from parsdic. The resulting dictionary, self.iparsdic, is what we will 
        # pass to method runtask.

        self.iparsdic = {**defaults, **self.argsdic}

    def printHelp(self):
        self.paramXmlInfo.printHelp()
    
    def run(self):
        """
        Method run

        If taskname is a Python module, therefore it is in the
        list pysaspkgs, then import it and pass to its run
        function the dictionary of parameters, iparsdic

        If taskname is not a Python SAS task, there will not be
        a run function, so we will invoke subprocess
        """
        self.logger.debug(f'Running {self.taskname}')
        self.logger.debug(f'Reading parameter file.')
        self.readparfile()
        self.logger.debug(f'Processing input arguments')
        self.processargs()
        if self.Exit:
            self.logger.debug(f'Exit called by processargs')
            return self.Exit
        sas_path = os.environ.get('SAS_PATH')

        if not sas_path:
            raise Exception('SAS_PATH is undefined! SAS not initialised?')

        pysaspkgs = []

        my_resources = importlib.resources.files("pysas")
        for line in (my_resources / "pysaspkgs").read_text().splitlines():
            pysaspkgs.append(line)

        # For Python based SAS tasks
        if self.taskname in pysaspkgs:
            self.logger.debug(f'Using Python based SAS task: {self.taskname}')
            # Add the environment options to iparsdic
            self.logger.debug(f'Adding environment options to iparsdic')
            self.iparsdic['options'] = dict(self.env_options)
            temp_dict = {}
            for k, v in self.env_options.items():
                match(k):
                    case 'SAS_CCFPATH' | 'SAS_CCF' | 'SAS_ODF' | 'SAS_VERBOSITY':
                        temp_dict[k] = os.getenv(k)
                        self.logger.debug(f'Storing previous setting of {k}={temp_dict[k]}')
                        os.environ[k] = v
                        self.logger.debug(f'Temporarely changing {k} to {v}')
                    case 'SAS_CLOBBER' | 'CCF_files' | 'WARNING' | 'TRACE':
                        # This needs to be handled by the individual task
                        pass
                    case _:
                        pass
            
            self.logger.debug(f'Importing module {self.taskname}')
            m = import_module('pysas.' + self.taskname + '.' + self.taskname)

            self.logger.debug(f'Running module {self.taskname}')
            m.run(self.iparsdic)
            
            # Reset Environment variables
            for k, v in temp_dict.items():
                match(k):
                    case 'SAS_CCFPATH' | 'SAS_CCF' | 'SAS_ODF' | 'SAS_VERBOSITY':
                        self.logger.debug(f'Resetting {k} to {temp_dict[k]}')
                        if temp_dict[k] is None:
                            os.environ.pop(k,None)
                        else:
                            os.environ[k] = temp_dict[k]
                    case 'SAS_CLOBBER' | 'CCF_files' | 'WARNING' | 'TRACE':
                        # This needs to be handled by the individual task
                        pass
                    case _:
                        pass
            
        # For all other SAS tasks
        else:
            # Build a list of parameters based on iparsdic
            self.logger.debug(f'Using non-Python based SAS task: {self.taskname}')
            cmd_list = []
            cmd_list.append(self.taskname)
            for k, v in self.iparsdic.items():
                if k == 'options':
                    cmd_list.append(v)
                    continue
                #Remove single quotes a double quotes from the python input parameters
                #SOC-SPR-7684
                if v.startswith("\"") or v.endswith("\""):
                    v = v.replace('"','')
                if v.startswith("'") or v.endswith("'"):
                    v = v.replace('\'','')
                
                if ' ' or '|' in v:
                    singparam = k + '=' + "'"
                    vc = v.split(' ')
                    for i in range(len(vc)):
                        if i == len(vc) - 1:
                            singparam += vc[i]
                        else:
                            singparam += vc[i] + ' '
                    singparam += "'"
                    cmd_list.append(singparam)
                else:
                    cmd_list.append(k + '=' + v)

            # Join all the parameters into a single command
            cmd = " ".join(cmd_list)
            if self.output_to_terminal:    
                print(f'Executing: \n{cmd}')

            try:
                logger = get_logger(self.taskname,
                                    toterminal  = self.output_to_terminal, 
                                    tofile      = self.output_to_file, 
                                    logfilename = self.logfilename,
                                    tasklogdir  = self.tasklogdir,
                                    pylogger    = False)
                                    # The only place pylogger should be set to false
                # Start the subprocess
                process = subprocess.Popen(cmd, 
                                           bufsize=1,
                                           shell=True,
                                           text=True,
                                           stdout=subprocess.PIPE,
                                           stderr=subprocess.STDOUT,
                                           universal_newlines=True)

                # Log stdout and stderr in real-time
                # For non-Python SAS tasks the stout and stderr are combined
                for line in process.stdout:
                    logger.info(f"{line.strip()}")

                # Wait for the process to complete and get the return code
                process.wait()

            except Exception as e:
                logger.exception(f"An error occurred while running the command: {e}")

            if process.returncode == 0:
                logger.success(f"{self.taskname} executed successfully!")
            else:
                logger.critical(f"{self.taskname} failed!")

    def runtask(self):
        """
        This method is here for legacy reasons since some Python
        based SAS tasks still create a MyTask object.
        e.g.
            # Instatntiate MyTask for the task
            t = MyTask('sasver', args)
            t.readparfile()
            t.processargs()
            t.runtask()
        """
        self.run()


class SASTask(MyTask):
    """
    Class SASTask is a child of MyTask and has access to all
    of the methods in MyTask. It's only purpose is to act as
    a placeholder for legacy code.
    """
    def __init__(self, taskname, inargs):
        super().__init__(taskname,inargs)
        warn(
             """
             The class SASTask has been depricated. Use MyTask instead.
             ex: from pysas.sastask import MyTask
             """)


# init_sas.py
#
# Written by: Ryan Tanner
# email: ryan.tanner@nasa.gov
# 
# This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
#
#    SAS is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    SAS is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with SAS. If not, see <http://www.gnu.org/licenses/>.

# init_sas.py

"""
init_sas.py

For auto initializing SAS on import of pySAS.

"""

# Standard library imports
import os, subprocess

# Third party imports

# Local application imports

# Function to initialize SAS

def add_environ_variable(variable,invalue,prepend=True):
        """
        variable (str) is the name of the environment variable to be set.
        
        value (str, or list) is the value to which the environment variable
        will be set.
        
        prepend (boolean) default=True, whether to prepend or append the 
        variable
        
        The function first checks if the enviroment variable already exists.
        If not it will be created and set to value.
        If variable alread exists the function will check if value is already
        present. If not it will add it either prepending (default) or appending.

        Returns
        -------
        None.

        """
        
        if isinstance(invalue, str):
            listvalue = [invalue]
        else:
            listvalue = invalue
            
        if not isinstance(listvalue, list):
            raise Exception('Input to add_environ_variable must be str or list!')
        
        for value in listvalue:
            environ_var = os.environ.get(variable)
            # Create variable if it does not exist.
            if not environ_var:
                os.environ[variable] = value
            else:
                splitpath = environ_var.split(os.pathsep)
                # Only add if the new value does not exist in the variable.
                if not value in splitpath:
                    if prepend:
                        splitpath.insert(0,value)
                    else:
                        splitpath.append(value)
                    os.environ[variable] = os.pathsep.join(splitpath)

def overwrite_environ_variable(variable,invalue):
    """
        variable (str) is the name of the environment variable to be set.
        
        value (str, or list) is the value to which the environment variable
        will be set.

        Will overwrite an environment variable to a new value. Will not check
        if variable exists or if it is defined.

        Returns
        -------
        None.
    """
    if isinstance(invalue, str):
        listvalue = [invalue]
    else:
        listvalue = invalue
        
    if not isinstance(listvalue, list):
        raise Exception('Input to overwrite_environ_variable must be str or list!')
    
    os.environ[variable] = os.pathsep.join(listvalue)

def initializesas(sas_dir, sas_ccfpath, verbosity = 4, suppress_warning = 1, image_viewer = 'ds9'):
    """
    Heasoft must be initialized first, separately.

    Inputs are:

        - sas_dir (required) directory where SAS is installed.

        - sas_ccfpath (required) directory where calibration files are located.

        - verbosity (optional, default = 4) SAS verbosity.

        - suppress_warning (optional, default = 1) 

    Returns:
    --------
    Information about SAS envirment veriables that were set.
    """

    ######
    # Checking LHEASOFT and inputs

    lheasoft = os.environ.get('LHEASOFT')
    if not lheasoft:
        raise Exception('LHEASOFT is not set. Please initialise HEASOFT')

    # Necessary to fix a problem with calls to heasoft from certain SAS functions.
    # add_environ_variable('HEADASNOQUERY','')
    # add_environ_variable('HEADASPROMPT','/dev/null')

    # Checks if a basic call to HEASOFT works.
    try:
        heasoft_test = subprocess.run(["fversion"],stdout = subprocess.DEVNULL)
    except FileNotFoundError:
        raise Exception('HEASOFT is not initialized. Please initialise HEASOFT')

    # Will not check the configuration file. Configuration file checking must happen 
    # in the calling script or function. sas_dir and sas_ccfpath must be passed in.
    if sas_dir is None:
        raise Exception('sas_dir must be provided to initialize SAS.')
    if sas_ccfpath is None:
        raise Exception('sas_ccfpath must be provided to initialize SAS.')

    # Hard overwrite. To set or reset values.
    overwrite_environ_variable('SAS_DIR',sas_dir)
    overwrite_environ_variable('SAS_CCFPATH',sas_ccfpath)
    overwrite_environ_variable('SAS_PATH',[sas_dir])
    
    binpath = [os.path.join(sas_dir,'bin'), os.path.join(sas_dir,'bin','devel')]
    libpath = [os.path.join(sas_dir,'lib'),os.path.join(sas_dir,'libextra'),os.path.join(sas_dir,'libsys')]
    perlpath = [os.path.join(sas_dir,'lib','perl5')]
    pythonpath = [os.path.join(sas_dir,'lib','python')]

    # Soft add. Will only add if value does not exist in enviroment variable.
    add_environ_variable('SAS_PATH',binpath+libpath+perlpath+pythonpath,prepend=False)
    add_environ_variable('PATH',binpath)
    add_environ_variable('LIBRARY_PATH',libpath,prepend=False)
    add_environ_variable('LD_LIBRARY_PATH',libpath,prepend=False)
    add_environ_variable('PERL5LIB',perlpath)
    add_environ_variable('PYTHONPATH',pythonpath)

    perllib = os.environ.get('PERLLIB')
    if perllib:
        splitpath = perllib.split(os.pathsep)
        add_environ_variable('PERL5LIB',splitpath,prepend=False)

    os.environ['SAS_VERBOSITY'] = '{}'.format(verbosity)
    os.environ['SAS_SUPPRESS_WARNING'] = '{}'.format(suppress_warning)
    os.environ['SAS_IMAGEVIEWER'] = '{}'.format(image_viewer)

    sas_path = os.environ.get('SAS_PATH')

    return_info = f"""
        SAS_DIR set to {sas_dir}
        SAS_CCFPATH set to {sas_ccfpath}
        SAS_PATH set to {sas_path}

        {libpath} added to LIBRARY_PATH
        {libpath} added to LD_LIBRARY_PATH
        {perlpath} added to PERL5LIB
        {pythonpath} added to PYTHONPATH
    """
    if perllib:
        return_info += f"""{perllib} added to PERLLIB"""
    return_info += f"""

    SAS_VERBOSITY set to {verbosity}
    SAS_SUPPRESS_WARNING set to {suppress_warning}
    """

    return return_info


# __init__.py for pysas
#
#

import os

# Check if SAS_DIR, SAS_PATH, and SAS_CCFPATH are already set.
sas_initialize = True
sas_ready      = False 
sas_dir        = os.environ.get('SAS_DIR')
sas_path       = os.environ.get('SAS_PATH')
sas_ccfpath    = os.environ.get('SAS_CCFPATH')

value = None

if sas_dir and \
   sas_ccfpath and \
   sas_path:
    if os.path.exists(sas_dir) and \
       os.path.exists(sas_ccfpath) and \
       os.path.exists(sas_path):
        # SAS_DIR, SAS_PATH, and SAS_CCFPATH enviroment variables already set.
        # SAS already initialized.
        sas_initialize = False
        sas_ready      = True
        # Check for SAS_VERBOSITY and SAS_SUPPRESS_WARNING
        # Set to default values if not present
        value = os.environ.get("SAS_VERBOSITY")
        if value is None:
            os.environ['SAS_VERBOSITY'] = '4'
        value = os.environ.get("SAS_SUPPRESS_WARNING")
        if value is None:
            os.environ['SAS_SUPPRESS_WARNING'] = '1'

# If SAS environment variables are not set look in the config file
from configparser import ConfigParser
from pathlib import Path

# Config class
class sas_config:
    def __init__(self, config_file=None):
        self.config = ConfigParser()

        # Raw defaults
        self.sas_cfg_defaults = {
            "suppress_warning" : 1,
            "verbosity"        : 4,
            "pysas_verbosity"  : "WARNING",
            "repo"             : "ESA",
            "work_dir_name"    : "work"
        }
        
        # Resolve the full path to the configuration file
        if config_file is None:
            home_dir = Path.home()
            # For SciServer
            if str(home_dir) == '/home/idies':
                user = os.environ.get('SCISERVER_USER_NAME')
                home_dir = home_dir / "workspace" / "Storage" / user / "persistent"
                if not home_dir.exists(): home_dir = Path.home()
            CONFIG_ROOT = os.environ.get("XDG_CONFIG_HOME", home_dir / ".config")
            CONFIG_ROOT = Path(CONFIG_ROOT).resolve()
            CONFIG_ROOT  = CONFIG_ROOT / 'sas'
            if not CONFIG_ROOT.exists():
                try:
                    os.makedirs(CONFIG_ROOT)
                except OSError:
                    #print(f'Unable to create config directory {CONFIG_ROOT}')
                    pass    
            config_file = CONFIG_ROOT / 'sas.cfg'
        else:
            config_file = Path(config_file).resolve()

        self.absolute_config_path = config_file

        if self.absolute_config_path.exists():
            # Read the configuration file
            self.config.read(self.absolute_config_path)
        else:
            self.config = ConfigParser(self.sas_cfg_defaults)
            if not self.config.has_section('sas'): self.config.add_section('sas')
            self.save_config()

    def get_setting(self, option, section = 'sas'):
        """
        Retrieves a setting from the configuration.
        """
        return self.config.get(section, option)

    def set_setting(self, option, value, section = 'sas'):
        """
        Sets a setting in the configuration.
        """
        if not self.config.has_section(section):
            self.config.add_section(section)
        self.config.set(section, option, value)

    def set_setting_and_save(self, option, value, section = 'sas'):
        """
        Sets a setting in the configuration.
        Saves the setting to file.
        """
        if not self.config.has_section(section):
            self.config.add_section(section)
        self.config.set(section, option, value)
        self.save_config()

    def save_config(self, config_file_path = None):
        """
        Saves the current configuration back to a file.
        """
        if config_file_path is None:
            absolute_config_path = self.absolute_config_path
        else:
            absolute_config_path = Path(config_file_path).resolve()
        try:
            with open(absolute_config_path, 'w') as file:
                self.config.write(file)
        except IOError:
            print(f'Unable to write config file to {absolute_config_path}')
            pass

    def show_current_config(self):
        """
        Shows the current configuration settings.
        """
        defaults = self.config.defaults()
        print(f'[{self.config.default_section}]')
        for k,v in defaults.items():
            print(f'{k} = {v}')
        
        for section in self.config.sections():
            print(f'\n[{section}]')
            for k,v in self.config.items(section):
                if k in defaults.keys():
                    if defaults[k] == v:
                        continue
                print(f'{k} = {v}')

    def show_config_file(self):
        """
        Prints the contents of the configuration file to the terminal.
        """
        with open(self.absolute_config_path, 'r') as file:
            content = file.read()
            print(content)

    def reset_to_defaults(self):
        """
        Resets config file to defaults.
        """
        self.config = ConfigParser(self.sas_cfg_defaults)
        if not self.config.has_section('sas'): self.config.add_section('sas')
        self.save_config()

    def simple_config(self, 
                      sas_dir = None, 
                      sas_ccfpath = None, 
                      data_dir = None,
                      repo = None):
        """
        For quick, simple configuration of pySAS.
        """
        home_dir = Path.home()
        if sas_dir is None: sas_dir = os.environ.get('SAS_DIR')
        if sas_ccfpath is None: sas_ccfpath = os.environ.get('SAS_CCFPATH')
        
        # For Fornax
        if str(sas_ccfpath) == '/opt/support-data/xmm_ccf':
            if repo is None:
                repo = 'fornax'
        
        # For SciServer
        if str(home_dir) == '/home/idies':
            if data_dir is None:
                user = os.environ.get('SCISERVER_USER_NAME')
                data_dir = os.path.join('/home/idies/workspace/Temporary/',user,'scratch/xmm_data')
            if repo is None:
                repo = 'sciserver'
        
        if data_dir is None: 
            data_dir = home_dir / 'xmm_data'
            data_dir = data_dir.resolve()

        if sas_dir: self.set_setting('sas_dir', sas_dir)
        if sas_ccfpath: self.set_setting('sas_ccfpath', sas_ccfpath)
        if repo: self.set_setting('repo', repo)
        self.set_setting('data_dir', str(data_dir))
        self.save_config()

# Create config object
sas_cfg = sas_config()

# Get configuration settings
# Checks if defaults from config file exist.
if sas_cfg.config.has_option('sas','sas_dir'):
    sas_dir_config = sas_cfg.get_setting("sas_dir")
    if not os.path.exists(sas_dir_config):
        print('There is a problem with SAS_DIR in the config file!')
        print(f'{sas_dir_config} does not exist!')
        print('Please set manually to initialize SAS.')
        sas_initialize = False
else:
    sas_initialize = False

if sas_cfg.config.has_option('sas','sas_ccfpath'):
    sas_ccfpath_config = sas_cfg.get_setting("sas_ccfpath")
    if not os.path.exists(sas_ccfpath_config):
        print('There is a problem with SAS_CCFPATH in the config file!')
        print(f'{sas_ccfpath_config} does not exist!')
        print('Please set manually to initialize SAS.')
        sas_initialize = False
else:
    sas_initialize = False

## For initialization of pySAS from configuration file.
#from .configutils import sas_cfg
from .init_sas import initializesas

# Initializes SAS if needed.
if sas_initialize:
    sas_verbosity        = sas_cfg.get_setting("verbosity")
    sas_suppress_warning = sas_cfg.get_setting("suppress_warning")
    sas_init_info = initializesas(sas_dir_config,
                                  sas_ccfpath_config,
                                  verbosity = sas_verbosity,
                                  suppress_warning = sas_suppress_warning)
    sas_ready = True


# Import pySAS modules
from . import sastask
from . import parser
from . import param
# from . import error
from . import init_sas
from . import sasutils
from . import config_pysas
from .version import VERSION, get_sas_version

# Get SAS version information
if sas_ready:
    return_list = get_sas_version()
else:
    return_list = ['NOT INITIALIZED','','','','','','']

SAS_RELEASE          = return_list[0]
SAS_AKA              = return_list[1]
SAS_COMPILATION_DATE = return_list[2]
SAS_COMPILATION_HOST = return_list[3]
SAS_COMPILATION_USER = return_list[4]
SAS_PLATFORM         = return_list[5]
SAS_COMMIT_ID        = return_list[6]

__version__ = f'pysas - (pysas-{VERSION}) [SAS-{SAS_RELEASE}]'

# Classes and functions needed at the top level
from .obsid.obsid import ObsID, PPSFiles
from .sastask import MyTask
from .print_version import print_sas_version
from .config_pysas import run_config
from .sasutils import download_data, generate_logger, update_calibration_files

# Will be depricated at some point
from .odfcontrol import odfcontrol

# API
__all__ = ['ObsID',
           'MyTask']

# Get rid of temporary variables to prevent possible conflicts.
if sas_initialize:
    del sas_dir_config, sas_ccfpath_config, sas_verbosity, sas_suppress_warning
del sas_initialize, sas_ready
del sas_dir, sas_path, sas_ccfpath
del value, return_list



# config_pysas.py
#
# Written by: Ryan Tanner
# email: ryan.tanner@nasa.gov
# 
# This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
#
#    SAS is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    SAS is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with SAS. If not, see <http://www.gnu.org/licenses/>.

# config_pysas.py

"""
    The purpose of this script is so that the user can set the pySAS
    defaults 
    
        sas_dir (required)
        sas_ccfpath (required)
        data_dir (optional)
        
    Once the defaults are set by the user, SAS will automatically be 
    initialized when pySAS is imported (import pysas).

    The user can also optionally set a default data directory (data_dir)
    where observation data files (odf) will be downloaded. A separate 
    subdirectory will be made for each observation ID (obsID).
    
    The default data directory can be set or change later using the 
    function set_sas_config_option().

    For example:

        from pysas import sas_cfg
        data_path = '/path/to/data/dir/'
        sas_cfg.set_setting_and_save('data_dir', data_path)
        
    The default values for the SAS directory (sas_dir), the path to the
    calibration files (sas_ccfpath), along with 'verbosity' and 
    'suppress_warning', can also be set in the same way.

    At any time the user can reset the config file to the defaults,

        from pysas import sas_cfg
        sas_cfg.reset_to_defaults()
"""

# Standard library imports
import os, glob

# Third party imports

# Local application imports
from pysas import sas_cfg
from pysas.init_sas import initializesas
from pysas.sasutils import update_calibration_files

__version__ = 'config_pysas (config_pysas-1.1)'

verbosity        = sas_cfg.get_setting('verbosity')
suppress_warning = sas_cfg.get_setting('suppress_warning')

def run_config():

    outcomment = """

    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        The purpose of this script is so the user can set the pySAS defaults 
        
            sas_dir (required)
            sas_ccfpath (required)
            data_dir (optional, recommended)
            
        Once the defaults are set by the user, SAS will automatically be 
        initialized when pySAS is imported (import pysas).

        The user can also optionally set a default data directory (data_dir)
        where observation data files (odf) will be downloaded. A separate 
        subdirectory will be made for each observation ID (obsID).
        
        The default data directory can be set or changed later using the 
        function set_setting_and_save().

        For example:

            from pysas import sas_cfg
            data_path = '/path/to/data/dir/'
            sas_cfg.set_setting_and_save('data_dir', data_path)
            
        The default values for the SAS directory (sas_dir), the path to the
        calibration files (sas_ccfpath), along with 'verbosity' and 
        'suppress_warning', can also be set in the same way.

        At any time the user can reset the config file to the defaults,

            from pysas import sas_cfg
            sas_cfg.reset_to_defaults()

    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    """

    print(outcomment)

    positive = ['y','yes','ye','yeah','yea','ys','aye','yup','totally','si','oui']
    negative = ['n','no','not','nay','no way','nine','non']
    esa = ['esa','e','es','europe']
    nasa = ['nasa','n','na','nas','ns','nsa','us','usa']

    ############## Getting sas_dir ##############
    
    # If sas_dir is already set
    psas_dir = os.environ.get('SAS_DIR')

    if psas_dir:
        print('Is this the correct SAS directory?')
        print('\n    {0}\n'.format(psas_dir))
        response = input('y/n: ')
        response = response.lower()
        if response in positive:
            sas_dir = psas_dir
            print(f'Setting SAS_DIR = {sas_dir}')
        elif response in negative:
            # Ask for SAS_DIR path
            scomment = '\nPlease provide the full path to the SAS install directory (SAS_DIR).\n'
            print(scomment)
            sas_dir = input('Full path to SAS: ')
        else:
            print(f'Your response, {response}, is not recognized.')
            print(f'Try any of these: {positive}')
            print(f'-or any of these: {negative}')
            raise Exception('Input not recognized!')
    else:
        # Ask for SAS_DIR path
        scomment = '\nPlease provide the full path to the SAS install directory (SAS_DIR).\n'
        print(scomment)
        sas_dir = input('Full path to SAS: ')

    if not sas_dir.startswith('/'): sas_dir = os.path.abspath(sas_dir)
    if sas_dir.endswith('/'): sas_dir = sas_dir[:-1]
    if not os.path.exists(sas_dir):
        print(f'SAS path {sas_dir} does not exist! Check path or SAS install!')
        raise Exception(f'SAS path {sas_dir} does not exist!')

    print('\n~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~')

    ############## Getting sas_ccfpath ##############

    psas_ccfpath = os.environ.get('SAS_CCFPATH')

    if psas_ccfpath:
        print('Is this the correct directory for the calibration files?')
        print('\n    {0}\n'.format(psas_ccfpath))
        response = input('y/n: ')
        response = response.lower()
        if response in positive:
            sas_ccfpath = psas_ccfpath
            print(f'Setting SAS_CCFPATH = {sas_ccfpath}')
        elif response in negative:
            # Ask for SAS_CCFPATH path
            scomment = '\nPlease provide the full path to the calibration files (SAS_CCFPATH).\n'
            print(scomment)
            sas_ccfpath = input('Full path to the calibration files: ')
        else:
            print(f'Your response, {response}, is not recognized.')
            print(f'Try any of these: {positive}')
            print(f'-or any of these: {negative}')
            raise Exception('Input not recognized!')
    else:
        scomment = """
            SAS_CCFPATH not set.

            Please provide the full path to the SAS calibration directory (SAS_CCFPATH).

        """
        print(scomment)
        sas_ccfpath = input('Full path to calibration files: ')
        if not sas_ccfpath.startswith('/'): sas_ccfpath = os.path.abspath(sas_ccfpath)
        if sas_ccfpath.endswith('/'): sas_ccfpath = sas_ccfpath[:-1]

    if not os.path.exists(sas_ccfpath):
        print(f'The directory {sas_ccfpath} was not found!')
        response = input('Should I create it? (y/n): ')
        response = response.lower()
        if response in positive:
            print(f'Creating: {sas_ccfpath}')
            os.mkdir(sas_ccfpath)
        elif response in negative:
            print('\nPlease create the directory for the calibration files!\n')
        else:
            print(f'Your response, {response}, is not recognized.')
            print(f'Try any of these: {positive}')
            print(f'-or any of these: {negative}')
            raise Exception('Input not recognized!')
            
    download_calibration = False
    esa_or_nasa = ''
    print('Would you like to download the current valid set of calibration files?\nWill download at the end of this script.')
    response2 = input('(y/n): ')
    response2 = response2.lower()
    if response2 in positive:
        download_calibration = True
        print('Which repository do you want to use to download the calibration files?')
        esa_or_nasa = input('ESA or NASA: ')
        esa_or_nasa = esa_or_nasa.lower()
        if esa_or_nasa in esa+nasa:
            pass
        else:
            print(f'Your response, {esa_or_nasa}, is not recognized.')
            print(f'Try any of these: {esa}')
            print(f'-or any of these: {nasa}')
            raise Exception('Input not recognized!')
    elif response2 in negative:
        print('Please make sure you download the calibration data!')
    else:
        print(f'Your response, {response2}, is not recognized.')
        print(f'Try any of these: {positive}')
        print(f'-or any of these: {negative}')
        raise Exception('Input not recognized!')
        
    print('\n~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~')

    ############## Getting data_dir ##############

    scomment = """
        No default data directory.

        Please provide the full path to the user data directory.
        (OPTIONAL, but recommended)

    """
    print(scomment)
    data_dir = input('Full path to user data directory: ')
    if not data_dir.startswith('/'): data_dir = os.path.abspath(data_dir)
    if data_dir.endswith('/'): data_dir = data_dir[:-1]
    print('\n~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~')

    # Check if data_dir exists. If not then create it.
    if not os.path.isdir(data_dir):
        print(f'{data_dir} does not exist. Creating it!')
        os.mkdir(data_dir)
        print(f'{data_dir} has been created!')
    else:
        print(f'\nData directory exists. Will use {data_dir} to download data.')

    # Check if paths for SAS_DIR and SAS_CCFPATH exist.
    if os.path.exists(sas_dir) and os.path.exists(sas_ccfpath):
        print('SAS_DIR and SAS_CCFPATH exist. Will use the following to initialize SAS:')
        print(f'     SAS_DIR = {sas_dir}')
        print(f'     SAS_CCFPATH = {sas_ccfpath}')
        initializesas(sas_dir, sas_ccfpath, verbosity=verbosity,suppress_warning=suppress_warning)

    if not os.path.exists(sas_dir):
        print(f'There is a problem with SAS_DIR {sas_dir}. Please check and try again.')
        raise Exception(f'There is a problem with SAS_DIR {sas_dir}. Please check and try again.')

    if not os.path.exists(sas_ccfpath) and not download_calibration:
        print(f'SAS_CCFPATH ({sas_ccfpath}) does not exist! Make sure you create it and download the calibration files!')
        

    # Set sas_dir in configuration settings
    sas_cfg.set_setting('sas_dir', sas_dir)

    # Set sas_ccfpath in configuration settings
    sas_cfg.set_setting('sas_ccfpath', sas_ccfpath)

    # Set data_dir in configuration settings
    sas_cfg.set_setting('data_dir', data_dir)

    # Save configuration settings to file
    sas_cfg.save_config()

    # Putting calibration file download in its own function since it is used in multiple locations.
    if download_calibration:
        result = update_calibration_files(repo=esa_or_nasa)

    scomment = f"""
        Success!

        SAS_DIR set to {sas_dir}
        SAS_CCFPATH set to {sas_ccfpath}
        data_dir set to {data_dir}

        Upon running the command 'import pysas' SAS will 
        automatically be initialized.

        The defaults can be changed at any time using the commands:

            from pysas import sas_cfg
            sas_cfg.set_setting_and_save('value_to_set',value)

        At any time the user can reset the config file to the defaults,

            from pysas import sas_cfg
            sas_cfg.reset_to_defaults()

    """
    print(scomment)



# configutils.py
#
# Written by: Ryan Tanner
# email: ryan.tanner@nasa.gov
# 
# This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
#
#    SAS is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    SAS is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with SAS. If not, see <http://www.gnu.org/licenses/>.

# configutils.py

# """
# configutils.py

# For configuring defaults of SAS enviroment variables.

# """

# # Standard library imports
# import os
# from configparser import ConfigParser

# # Third party imports

# # Local application imports

# __version__ = 'configutils (configutils-1.0)'

# ###############
# # Configuration

# sas_cfg_defaults = {
#     "sas_dir": "/does/not/exist",
#     "sas_ccfpath": "/does/not/exist",
#     "data_dir": "/does/not/exist",
#     "verbosity": 4,
#     "pysas_verbosity": "WARNING",
#     "suppress_warning": 1
# }

# CURRENT_CONFIG_FILE = None

# if (os.path.expanduser("~") == '/home/idies') and \
#     os.path.exists('/home/idies/workspace/Storage/'):
#     # If on SciServer, get enviroment variables. Replace defaults.
#     sas_dir     = os.environ.get('SAS_DIR')
#     sas_ccfpath = os.environ.get('SAS_CCFPATH')
#     sas_cfg_defaults['sas_dir']     = sas_dir
#     sas_cfg_defaults['sas_ccfpath'] = sas_ccfpath
#     sas_cfg = ConfigParser(sas_cfg_defaults)
#     sas_cfg.add_section("sas")
# else:
#     # If not on SciServer.
#     # Check if .config directory exists. If not, make it.
#     config_root = os.environ.get(
#         "XDG_CONFIG_HOME", os.path.join(os.environ.get('HOME'), ".config")
#     )
#     CONFIG_DIR = os.path.join(config_root, "sas")
#     if not os.path.exists(CONFIG_DIR):
#         try:
#             os.makedirs(CONFIG_DIR)
#         except OSError:
#             raise Exception( f'Unable to make SAS config directory: {CONFIG_DIR}')

#     # Check if sas.cfg file exists. If not, make it. Populate with defaults.
#     CURRENT_CONFIG_FILE = os.path.join(CONFIG_DIR, "sas.cfg")
#     if not os.path.exists(CURRENT_CONFIG_FILE):
#         cp = ConfigParser(sas_cfg_defaults)
#         cp.add_section("sas")
#         try:
#             with open(CURRENT_CONFIG_FILE, "w") as new_cfg:
#                 cp.write(new_cfg)
#         except IOError:
#             raise Exception( f'Unable to write to SAS config file: {CURRENT_CONFIG_FILE}')

#     sas_cfg = ConfigParser(sas_cfg_defaults)
#     sas_cfg.read([CURRENT_CONFIG_FILE, "sas.cfg"])
#     if not sas_cfg.has_section("sas"): sas_cfg.add_section("sas")    

# ######### Functions #########

# def set_sas_config_option(option, value):
#     """
#     Set default SAS configuration values.

#     This sets values as default for future sessions.

#     Parameters
#     ----------
#     option : string
#         The option to change.
#     value : number or string
#         The value to set the option to.
#     """
#     if os.path.exists(CURRENT_CONFIG_FILE):
#         option = option.lower()
#         sas_cfg.set("sas", option, value=str(value))
#         with open(CURRENT_CONFIG_FILE, "w") as new_cfg:
#             sas_cfg.write(new_cfg)
#     else:
#         print('No SAS configuration file found! Cannot set default value for:')
#         print('Option: {0} ; Value: {1}'.format(option,value))

# def clear_sas_defaults():
#     """
#     Clears all SAS defaults set by user. User will have to reset SAS defaults.
#             sas_dir
#             sas_ccfpath
#             data_dir
#             verbosity
#             suppress_warning
#     """
#     if os.path.exists(CURRENT_CONFIG_FILE):
#         os.remove(CURRENT_CONFIG_FILE)
#     else:
#         print('SAS configuration file, {0}, not found! Cannot remove!'.format(CURRENT_CONFIG_FILE))




# setup_pysas.py
#
# Written by: Ryan Tanner
# email: ryan.tanner@nasa.gov
# 
# This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
#
#    SAS is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    SAS is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with SAS. If not, see <http://www.gnu.org/licenses/>.

# setup_pysas.py

"""
    Wrapper script that points to config_pysas
"""
import pysas

pysas.config_pysas.run_config()


#!/usr/bin/env python

#	ESA (C) 2000-2018 
#	This file is part of ESA's XMM-Newton Scientific Analysis System
#	(SAS).
#
#	SAS is free software: you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation, either version 3 of the License, or
#	(at your option) any later version.
#
#	SAS is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#	   
#	You should have received a copy of the GNU General Public License
#	along with SAS.  If not, see <http://www.gnu.org/licenses/>.
import sys
 
from PyQt5.QtWidgets import QApplication, QMainWindow, QMenu, QVBoxLayout, QSizePolicy, QMessageBox, QWidget, QPushButton
from PyQt5.QtGui import QIcon
 
from PyQt5.QtGui import QIcon, QPixmap
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.figure import Figure
import matplotlib.pyplot as plt
from astropy.visualization import astropy_mpl_style
from astropy.io import fits
from matplotlib.colors import LogNorm
import random

 
 
class App(QMainWindow):
 
    def __init__(self):
        super().__init__()
        self.left = 10
        self.top = 10
        self.title = 'XMM-Newton image'
        self.width = 500
        self.height = 500
        self.initUI()
 
    def initUI(self):
        self.setWindowTitle(self.title)
        self.setGeometry(self.left, self.top, self.width, self.height)
 
        m = PlotCanvas(self, width=5, height=5)
        m.move(0,0)       
 
        self.show()
 
 
class PlotCanvas(FigureCanvas):
 
    def __init__(self, parent=None, width=6, height=6, dpi=100):
        fig = Figure(figsize=(width, height), dpi=dpi)
        self.axes = fig.add_subplot(111)
 
        FigureCanvas.__init__(self, fig)
        self.setParent(parent)
 
        FigureCanvas.setSizePolicy(self,
                QSizePolicy.Expanding,
                QSizePolicy.Expanding)
        FigureCanvas.updateGeometry(self)
        self.plot()
 
 
    def plot(self):
        
        #image_file="/home/aibarra/pru/PythonWorkspace/xmmextractorTest/images/m2_S003_ImagingEvts_image_full.fits"
        image_file=""
        fits.info(image_file)

        image_data = fits.getdata(image_file, ext=0)

        ax = self.figure.add_subplot(111)
        ax.imshow(image_data,cmap='hot', norm=LogNorm())                    
        ax.set_title('MOS-1 Example')
        self.draw()
        
        
        
 
if __name__ == '__main__':
    app = QApplication(sys.argv)
    ex = App()
    sys.exit(app.exec_())
